(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[     62961,       1789]
NotebookOptionsPosition[     20009,        786]
NotebookOutlinePosition[     58160,       1624]
CellTagsIndexPosition[     57990,       1617]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["Using Global variables", "Title"],

Cell[" Code Generation using MathCode", "Subtitle"],

Cell["\<\
Copyright \[Copyright] MathCore AB 1998-2004\
\>", "Copyright"],

Cell[CellGroupData[{

Cell["How to use global variables", "Section"],

Cell[CellGroupData[{

Cell[TextData[{
 "Load the MathCode system into ",
 StyleBox["Mathematica",
  FontSlant->"Italic"]
}], "Subsection"],

Cell[CellGroupData[{

Cell[TextData[{
 "Needs[\"",
 StyleBox["MathCode`",
  FontColor->RGBColor[0, 0, 1]],
 "\"]"
}], "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"MathCode C++ 1.4.4\"\>", "\[InvisibleSpace]", "\<\" for \"\>", 
   "\[InvisibleSpace]", "\<\"vc100\"\>", 
   "\[InvisibleSpace]", "\<\" loaded from \"\>", 
   "\[InvisibleSpace]", "\<\"C:\\\\b\\\\omath\\\\MathCode\"\>"}],
  SequenceForm[
  "MathCode C++ 1.4.4", " for ", "vc100", " loaded from ", 
   "C:\\b\\omath\\MathCode"],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660053226003*^9}]
}, Open  ]],

Cell["This is needed if many notebooks run in a sequence", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"SetCurrentPackage", "[", "\"\<Global\>\"", "]"}], " "}]], "Input"],

Cell[BoxData["\<\"Global\"\>"], "Output",
 CellChangeTimes->{3.5086600532885027`*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Set Working Directory", "Subsection"],

Cell["SetDirectory[$MCRoot<>\"/Demos/GlobalVariable\"];", "Input"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Define a simple function", "Subsection"],

Cell["Typed definition", "Text"],

Cell["foo[Integer n_] ->  Real := Sin[1/n]", "Input",
 CellChangeTimes->{3.5086601711619935`*^9}],

Cell[TextData[{
 "Evaluation in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"]
}], "Text"],

Cell[CellGroupData[{

Cell["foo[2]", "Input"],

Cell[BoxData[
 RowBox[{"Sin", "[", 
  FractionBox["1", "2"], "]"}]], "Output",
 CellChangeTimes->{3.5086600533353767`*^9}]
}, Open  ]],

Cell[TextData[{
 StyleBox["foo",
  FontFamily->"Courier New"],
 " is still a normal ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " function"
}], "Text"],

Cell[CellGroupData[{

Cell["??foo", "Input"],

Cell["Global`foo", "Print", "PrintUsage",
 CellChangeTimes->{3.508660053429126*^9},
 CellTags->"Info3508663653-8109590"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"foo", "[", "n_", "]"}], ":=", 
         RowBox[{"Sin", "[", 
          FractionBox["1", "n"], "]"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`foo],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086600534447503`*^9},
 CellTags->"Info3508663653-8109590"]
}, Open  ]],

Cell["\<\
the type information is stored somewhere else and is only used for the code \
generation\
\>", "Text"],

Cell[CellGroupData[{

Cell["getDefinition[foo]", "Input"],

Cell[BoxData[
 RowBox[{"Hold", "[", 
  RowBox[{
   RowBox[{
    RowBox[{"(", 
     RowBox[{
      RowBox[{"{", "Integer", "}"}], "\[Rule]", "Real"}], ")"}], " ", 
    RowBox[{"foo", "[", "n_", "]"}]}], ":=", 
   RowBox[{"Sin", "[", 
    FractionBox["1", "n"], "]"}]}], "]"}]], "Output",
 CellChangeTimes->{3.508660053491625*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A Global Matrix", "Subsection"],

Cell["Define a matrix", "Text"],

Cell["\<\
Declare[
\tReal[3,3] xymatrix 
\t]\
\>", "Input"],

Cell["\<\
This matrix definition will also create/initialize the matrix. This to make, \
e.g, xymatrix[[2,3]] valid directly.\
\>", "Text"],

Cell[CellGroupData[{

Cell["xymatrix", "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "999.`"}], ",", 
     RowBox[{"-", "999.`"}], ",", 
     RowBox[{"-", "999.`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "999.`"}], ",", 
     RowBox[{"-", "999.`"}], ",", 
     RowBox[{"-", "999.`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "999.`"}], ",", 
     RowBox[{"-", "999.`"}], ",", 
     RowBox[{"-", "999.`"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5086600535228744`*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Matrix Function", "Subsection"],

Cell["Initializer", "Text"],

Cell["xyinit[Real x_] -> {} := xymatrix = Table[x,{3},{3}];", "Input"],

Cell["Updates the matrix", "Text"],

Cell["\<\
fee[Real x_] -> Real[_,_] := 
\tModule[ { Real[3,3] m },
\t    m = Table[foo[i+j],{i,3},{j,3}]; 
\t\txymatrix = xymatrix + m;
\t\txymatrix
]\
\>", "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"xyinit", "[", "0.", "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.508660053554124*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["fee[1]", "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "0.479425538604203`", ",", "0.3271946967961522`", ",", 
     "0.24740395925452294`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.3271946967961522`", ",", "0.24740395925452294`", ",", 
     "0.19866933079506122`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.24740395925452294`", ",", "0.19866933079506122`", ",", 
     "0.16589613269341502`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.508660053569749*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["fee[1]", "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "0.958851077208406`", ",", "0.6543893935923044`", ",", 
     "0.4948079185090459`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.6543893935923044`", ",", "0.4948079185090459`", ",", 
     "0.39733866159012243`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.4948079185090459`", ",", "0.39733866159012243`", ",", 
     "0.33179226538683004`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.508660053585374*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["fee[1]", "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "1.438276615812609`", ",", "0.9815840903884566`", ",", 
     "0.7422118777635688`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.9815840903884566`", ",", "0.7422118777635688`", ",", 
     "0.5960079923851836`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.7422118777635688`", ",", "0.5960079923851836`", ",", 
     "0.49768839808024506`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.508660053616623*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Generate C++ and Build Executables", "Subsection"],

Cell[CellGroupData[{

Cell["BuildCode[]", "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 3 function(s)\"\>"], "Print",
 CellChangeTimes->{3.508660055991593*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Install the Mathlink Executable", "Subsection"],

Cell[CellGroupData[{

Cell["InstallCode[];", "Input"],

Cell[BoxData[
 InterpretationBox[
  RowBox[{"\<\"Global\"\>", "\[InvisibleSpace]", "\<\" is installed.\"\>"}],
  SequenceForm["Global", " is installed."],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086600596634207`*^9}]
}, Open  ]],

Cell["Run some examples", "Text"],

Cell["xyinit[1.]", "Input"],

Cell[CellGroupData[{

Cell["fee[2.]", "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "1.479425538604203`", ",", "1.3271946967961523`", ",", 
     "1.247403959254523`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "1.3271946967961523`", ",", "1.247403959254523`", ",", 
     "1.1986693307950613`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "1.247403959254523`", ",", "1.1986693307950613`", ",", 
     "1.165896132693415`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5086600597259197`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell["fee[2.]", "Input"],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{
    "1.958851077208406`", ",", "1.6543893935923046`", ",", 
     "1.494807918509046`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "1.6543893935923046`", ",", "1.494807918509046`", ",", 
     "1.3973386615901227`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "1.494807918509046`", ",", "1.3973386615901227`", ",", 
     "1.33179226538683`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5086600597415447`*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Switch between ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " and Externals"
}], "Subsection"],

Cell["The different definitions (DownValues) are stored.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"??", "foo"}]], "Input"],

Cell["Global`foo", "Print", "PrintUsage",
 CellChangeTimes->{3.508660059835294*^9},
 CellTags->"Info3508663659-8109590"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"ExternalDownValues", "[", "foo", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"HoldPattern", "[", 
            RowBox[{"foo", "[", "n_Integer", "]"}], "]"}], "\[RuleDelayed]", 
           RowBox[{"ExternalCall", "[", 
            RowBox[{
             RowBox[{"LinkObject", "[", 
              
              RowBox[{"\<\"\\\"C:\\\\b\\\\omath\\\\MathCode\\\\Demos\\\\\
GlobalVariable\\\\Globalml.exe\\\"\"\>", ",", "15", ",", "8"}], "]"}], ",", 
             RowBox[{"CallPacket", "[", 
              RowBox[{"1", ",", 
               RowBox[{"{", "n", "}"}]}], "]"}]}], "]"}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SourceDownValues", "[", "foo", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"HoldPattern", "[", 
            RowBox[{"foo", "[", "n_", "]"}], "]"}], "\[RuleDelayed]", 
           RowBox[{"Sin", "[", 
            FractionBox["1", "n"], "]"}]}], "}"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]},
    {" "},
    {GridBox[{
       {
        RowBox[{
         RowBox[{"foo", "[", "n_Integer", "]"}], ":=", 
         RowBox[{"ExternalCall", "[", 
          RowBox[{
           RowBox[{"LinkObject", "[", 
            
            RowBox[{"\<\"\\\"C:\\\\b\\\\omath\\\\MathCode\\\\Demos\\\\\
GlobalVariable\\\\Globalml.exe\\\"\"\>", ",", "15", ",", "8"}], "]"}], ",", 
           RowBox[{"CallPacket", "[", 
            RowBox[{"1", ",", 
             RowBox[{"{", "n", "}"}]}], "]"}]}], "]"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`foo],
  Editable->False]], "Print",
 CellChangeTimes->{3.5086600598509183`*^9},
 CellTags->"Info3508663659-8109590"]
}, Open  ]],

Cell[TextData[{
 "To use the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " source definition instead"
}], "Text"],

Cell[CellGroupData[{

Cell["DeactivateCode[foo]", "Input"],

Cell[BoxData[
 TagBox["foo",
  HoldForm]], "Output",
 CellChangeTimes->{3.508660059897793*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"??", "foo"}]], "Input"],

Cell["Global`foo", "Print", "PrintUsage",
 CellChangeTimes->{3.5086600600071664`*^9},
 CellTags->"Info3508663659-8109590"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"ExternalDownValues", "[", "foo", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"HoldPattern", "[", 
            RowBox[{"foo", "[", "n_Integer", "]"}], "]"}], "\[RuleDelayed]", 
           RowBox[{"ExternalCall", "[", 
            RowBox[{
             RowBox[{"LinkObject", "[", 
              
              RowBox[{"\<\"\\\"C:\\\\b\\\\omath\\\\MathCode\\\\Demos\\\\\
GlobalVariable\\\\Globalml.exe\\\"\"\>", ",", "15", ",", "8"}], "]"}], ",", 
             RowBox[{"CallPacket", "[", 
              RowBox[{"1", ",", 
               RowBox[{"{", "n", "}"}]}], "]"}]}], "]"}]}], "}"}]}]},
       {" "},
       {
        RowBox[{
         RowBox[{"SourceDownValues", "[", "foo", "]"}], "^=", 
         RowBox[{"{", 
          RowBox[{
           RowBox[{"HoldPattern", "[", 
            RowBox[{"foo", "[", "n_", "]"}], "]"}], "\[RuleDelayed]", 
           RowBox[{"Sin", "[", 
            FractionBox["1", "n"], "]"}]}], "}"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]},
    {" "},
    {GridBox[{
       {
        RowBox[{
         RowBox[{"foo", "[", "n_", "]"}], ":=", 
         RowBox[{"Sin", "[", 
          FractionBox["1", "n"], "]"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`foo],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660060022791*^9},
 CellTags->"Info3508663659-8109590"]
}, Open  ]],

Cell["Uninstall the code", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"UninstallCode", "[", "]"}], ";"}]], "Input"],

Cell[TextData[{
 "which will clean up things and terminate the ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " process."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"??", "foo"}]], "Input"],

Cell["Global`foo", "Print", "PrintUsage",
 CellChangeTimes->{3.5086600601634145`*^9},
 CellTags->"Info3508663660-8109590"],

Cell[BoxData[
 InterpretationBox[GridBox[{
    {GridBox[{
       {
        RowBox[{
         RowBox[{"foo", "[", "n_", "]"}], ":=", 
         RowBox[{"Sin", "[", 
          FractionBox["1", "n"], "]"}]}]}
      },
      BaselinePosition->{Baseline, {1, 1}},
      GridBoxAlignment->{
       "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
        "RowsIndexed" -> {}},
      GridBoxItemSize->{"Columns" -> {{
           Scaled[0.999]}}, "ColumnsIndexed" -> {}, "Rows" -> {{1.}}, 
        "RowsIndexed" -> {}}]}
   },
   BaselinePosition->{Baseline, {1, 1}},
   GridBoxAlignment->{
    "Columns" -> {{Left}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
     "RowsIndexed" -> {}}],
  Definition[$CellContext`foo],
  Editable->False]], "Print",
 CellChangeTimes->{3.508660060179039*^9},
 CellTags->"Info3508663660-8109590"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["A glance at the Generated Code", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<Global.cc\>\"", "]"}]], "Input"],

Cell["\<\
#include \"Global.h\"

#include \"Global.icc\"

#include <math.h>
doubleNN Global_Tfee ( const double &x)
{
    doubleNN m(3, 3);
    int j_D1521_D1524;
    int i_D1520_D1526;

    m.SetShape(Max(3,0), Max(3,0) );
    i_D1520_D1526 = 1;
    while (i_D1520_D1526 <= 3)
    {
        j_D1521_D1524 = 1;
        while (j_D1521_D1524 <= 3)
        {
            m(i_D1520_D1526, j_D1521_D1524) = Global_Tfoo (i_D1520_D1526+
                j_D1521_D1524);
            j_D1521_D1524 = j_D1521_D1524+1;
        }
        i_D1520_D1526 = i_D1520_D1526+1;
    }
    Global_Txymatrix = Global_Txymatrix+m;
    return Global_Txymatrix;
}

double Global_Tfoo ( const int &n)
{
    return sin(double (1)/n);
}

void Global_TGlobalInit ()
{
; 
}

void Global_Txyinit ( const double &x)
{
    int _D4_D1639_D1642;
    int _D3_D1638_D1644;
    Global_Txymatrix.SetShape(Max(3,0), Max(3,0) );
    _D3_D1638_D1644 = 1;
    while (_D3_D1638_D1644 <= 3)
    {
        _D4_D1639_D1642 = 1;
        while (_D4_D1639_D1642 <= 3)
        {
            Global_Txymatrix(_D3_D1638_D1644, _D4_D1639_D1642) = x;
            _D4_D1639_D1642 = _D4_D1639_D1642+1;
        }
        _D3_D1638_D1644 = _D3_D1638_D1644+1;
    }; 
}

doubleNN Global_Txymatrix(3, 3);\
\>", "Print",
 CellChangeTimes->{3.5086600602259135`*^9}]
}, Open  ]],

Cell[TextData[{
 "This code is used to produce two executables.\nOne executable for the ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " connection"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FileNames", "[", "\"\<*ml.exe\>\"", "]"}]], "Input"],

Cell[BoxData[
 RowBox[{"{", "\<\"Globalml.exe\"\>", "}"}]], "Output",
 CellChangeTimes->{3.508660060241538*^9}]
}, Open  ]],

Cell["\<\
The main function for the standalone executable can be given to BuildCode as \
an option.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
CompilePackage[
\tMainFileAndFunction ->
\t\t\"int main() {
\t\t\tGlobal_Txyinit(1.0);
\t\t\tcout << Global_Tfee(1.0);
\t\t\treturn (0);
\t\t}\"]\
\>", "Input"],

Cell[BoxData["\<\"Successful compilation to C++: 3 function(s)\"\>"], "Print",
 CellChangeTimes->{3.5086600626790075`*^9}]
}, Open  ]],

Cell["MakeBinary[StandAloneExecutable -> True]", "Input"],

Cell[CellGroupData[{

Cell["FileNames[\"Global.exe\"]", "Input"],

Cell[BoxData[
 RowBox[{"{", "\<\"Global.exe\"\>", "}"}]], "Output",
 CellChangeTimes->{3.5086600657102184`*^9}]
}, Open  ]],

Cell["Run the stand alone executable", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", 
   RowBox[{"Run", "[", "\"\<./Global.exe > tmp\>\"", "]"}], "*)"}], " ", 
  RowBox[{"(*", 
   RowBox[{"for", " ", "linux", " ", "version"}], "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{"Run", "[", "\"\<Global.exe > tmp\>\"", "]"}], " ", 
  RowBox[{"(*", 
   RowBox[{"for", " ", "windows", " ", "version"}], "*)"}]}]], "Input"],

Cell[BoxData["0"], "Output",
 CellChangeTimes->{3.5086600658352165`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"FilePrint", "[", "\"\<tmp\>\"", "]"}]], "Input"],

Cell["\<\
  1.47943 1.32719 1.2474    
  1.32719 1.2474 1.19867    
  1.2474 1.19867 1.1659      \
\>", "Print",
 CellChangeTimes->{3.5086600658664665`*^9}]
}, Open  ]],

Cell["Clean the directory", "Text"],

Cell["CleanMathCodeFiles[Confirm -> False, CleanAllBut -> {}];", "Input"]
}, Open  ]]
}, Open  ]]
},
ScreenStyleEnvironment->"Working",
WindowToolbars->{},
WindowSize->{678, 969},
WindowMargins->{{90, Automatic}, {Automatic, 0}},
ShowSelection->True,
ShowCellTags->True,
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (April 20, 2007)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Subtitle"], 
      Cell[
      "\nModify the definitions below to change the default appearance of all \
cells in \na given style.  Make modifications to any definition using \
commands in the \nFormat menu.\n", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, ScriptMinSize -> 
          9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          ScriptMinSize -> 12, FontSize -> 16], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {"Margins" -> {1, 1}, "Widths" -> {0, 5}}, 
          ScriptMinSize -> 8, FontSize -> 11], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, ScriptMinSize -> 
          7, FontSize -> 10, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "\nThe options defined for the style below will be used at the \
Notebook level.\n", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, 
          PageHeaderLines -> {True, True}, 
          PrintingOptions -> {
           "FacingPages" -> True, "FirstPageHeader" -> False}, 
          CellFrameLabelMargins -> 6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 0}, CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontSize -> 
             34, FontColor -> GrayLevel[1], Background -> RGBColor[0, 0, 1]], 
            
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 0}, FontSize -> 
             44], 
            Cell[
             StyleData["Title", "Condensed"], CellMargins -> {{0, 0}, {0, 0}},
              FontSize -> 20], 
            Cell[
             StyleData["Title", "Printout"], CellMargins -> {{0, 0}, {0, 0}}, 
             FontSize -> 24, FontTracking -> "Plain", Background -> 
             GrayLevel[0]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 3}, ParagraphIndent -> -96, CounterIncrements -> 
             "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 18, FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 6}, 
             ParagraphIndent -> -157, FontSize -> 30], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -78, 
             FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -85, 
             FontSize -> 16, Background -> GrayLevel[0.6]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {30, 10}}, PageBreakBelow -> False, 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 14, 
             FontSlant -> "Italic"], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{8, 10}, {40, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{9, 10}, {50, 10}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{6, 0}, {0, 1}}, CellDingbat -> 
             None, CellMargins -> {{12, Inherited}, {4, 24}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 6, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 16, FontWeight -> "Bold", 
             FontColor -> RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{10, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, FontSize -> 14, FontTracking -> 
             "Plain", FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], 
             CellMargins -> {{12, Inherited}, {8, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{11, 10}, {8, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {4, 40}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> "\[FilledSquare]", 
             CellMargins -> {{25, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{29, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{22, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{21, 0}, {4, 20}}, FontSize -> 11]}, Open]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Text", FontFamily -> "Times"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {4, 4}},
              LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{9, 0}, {4, 4}}, 
             ParagraphSpacing -> {0, 6}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 6}, 
             CounterIncrements -> "SmallText", FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {2, 2}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{9, 0}, {4, 4}}, FontSize -> 7]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "\nThe cells in this section define styles used for input and output \
to the \nkernel.  Be careful when modifying, renaming, or removing these \
styles, \nbecause the front end associates special meanings with these style \
names.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             CounterIncrements -> "Input", FontWeight -> "Bold", Background -> 
             RGBColor[1, 1, 0]], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{44, 0}, {6, 6}},
              Background -> GrayLevel[0.8]]}, Closed]], 
         Cell[
          StyleData["InlineInput"], Evaluatable -> True, CellGroupingRules -> 
          "InputGrouping", CellHorizontalScrolling -> True, PageBreakWithin -> 
          False, GroupPageBreakWithin -> False, DefaultFormatType -> 
          DefaultInputFormatType, AutoItalicWords -> {}, FormatType -> 
          InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
          CounterIncrements -> "Input", FontWeight -> "Bold"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, CellEditDuplicate -> True, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, 
             CellLabelMargins -> {{3, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Output", Background -> 
             RGBColor[1, 0.900008, 0.900008]], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{62, Inherited}, {12, 5}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{40, Inherited}, {4, 1}}], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{44, 0}, {6, 2}}, Background -> 
             GrayLevel[0.900008]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontColor -> RGBColor[1, 0, 0]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Graphics", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             CellMargins -> {{52, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {175, 175}], 
            Cell[
             StyleData["Graphics", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {250, 250}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 11, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 8], 
            Cell[
             StyleData["CellLabel", "Printout"], FontSize -> 8, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Unique Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {2, 10}}, FontSize -> 16, FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["Author", "Presentation"], 
             CellMargins -> {{12, 10}, {2, 12}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}, FontSize -> 20], 
            Cell[
             StyleData["Author", "Condensed"], 
             CellMargins -> {{8, 10}, {1, 4}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}, FontSize -> 12], 
            Cell[
             StyleData["Author", "Printout"], 
             CellMargins -> {{9, 0}, {4, 12}}, ParagraphSpacing -> {0, 6}, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Copyright"], ShowCellBracket -> False, 
             CellMargins -> {{10, 10}, {40, 2}}, FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Presentation"], 
             CellMargins -> {{12, 10}, {50, 2}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["Copyright", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 1}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Printout"], 
             CellMargins -> {{9, 0}, {72, 4}}, FontSize -> 7]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Example"], CellMargins -> {{12, 10}, {5, 12}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Example", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Example", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 20}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Example", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Example", "Printout"], 
             CellMargins -> {{9, 0}, {4, 10}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Definition"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Definition", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Definition", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Definition", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Definition", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.6]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Theorem"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Theorem", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Theorem", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Theorem", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Theorem", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.4]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseMain"], CellFrame -> {{6, 0}, {0, 1}}, 
             CellMargins -> {{12, Inherited}, {4, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameColor -> RGBColor[0.571389, 0.19675, 0.570504], 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 16, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["ExerciseMain", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["ExerciseMain", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["ExerciseMain", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, CellFrameColor -> 
             GrayLevel[0.500008], FontSize -> 14, FontTracking -> "Plain", 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Exercise"], CellDingbat -> "\[FilledDownTriangle]", 
             CellMargins -> {{23, Inherited}, {4, 18}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Exercise", "Presentation"], 
             CellMargins -> {{33, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Exercise", "Condensed"], 
             CellMargins -> {{17, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Exercise", "Printout"], 
             CellFrame -> {{0, 0}, {0.5, 0}}, CellDingbat -> None, 
             CellMargins -> {{9, 0}, {6, 20}}, FontSize -> 11, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseText"], CellMargins -> {{62, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 8}, 
             CounterIncrements -> "ExcersiceText", FontFamily -> "Times"], 
            Cell[
             StyleData["ExerciseText", "Presentation"], 
             CellMargins -> {{74, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["ExerciseText", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 2}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["ExerciseText", "Printout"], 
             CellMargins -> {{54, 0}, {4, 4}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Itemize"], CellDingbat -> "\[FilledSmallCircle]", 
             CellMargins -> {{65, Inherited}, {8, 8}}, FontFamily -> "Times", 
             FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Itemize", "Presentation"], 
             CellMargins -> {{Inherited, 10}, {8, 12}}, LineSpacing -> {1, 0},
              FontSize -> 18], 
            Cell[
             StyleData["Itemize", "Condensed"], 
             CellMargins -> {{Inherited, Inherited}, {2, 12}}, FontSize -> 
             10], 
            Cell[
             StyleData["Itemize", "Printout"], 
             CellMargins -> {{Inherited, 0}, {4, 20}}, FontSize -> 11]}, 
           Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayCode"], CellFrame -> 2, 
             CellMargins -> {{52, 10}, {8, 8}}, CellHorizontalScrolling -> 
             True, PageBreakWithin -> True, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             FontWeight -> "Bold", Background -> RGBColor[1, 1, 0]], 
            Cell[
             StyleData["DisplayCode", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["DisplayCode", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["DisplayCode", "Printout"], 
             CellMargins -> {{44, 0}, {6, 6}}, Background -> GrayLevel[0.8]]},
            Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], CellMargins -> {{62, 10}, {2, 10}}, 
             CellHorizontalScrolling -> True, ScriptLevel -> 0, 
             SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 18], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["DisplayFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ChemicalFormula"], CellMargins -> {{62, 10}, {2, 10}},
              AutoSpacing -> False, ScriptBaselineShifts -> {0.6, Automatic}, 
             ScriptLevel -> 1, SingleLetterItalics -> False, ZeroWidthTimes -> 
             True], 
            Cell[
             StyleData["ChemicalFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellMargins -> {{12, 10}, {Inherited, 6}}, 
             FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{13, 30}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{9, 0}, {Inherited, 4}}, FontSize -> 9.5]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Automatic Numbering", "Section"], 
         Cell[
         "\nThe following styles are useful for numbered equations, figures, \
etc.  They \nautomatically give the cell a FrameLabel containing a reference \
to a \nparticular counter, and also increment that counter.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedEquation"], 
             CellMargins -> {{62, 10}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, 
                Cell[
                 TextData[{"(", 
                   CounterBox["NumberedEquation"], ")"}]]}, {None, None}}, 
             DefaultFormatType -> DefaultInputFormatType, CounterIncrements -> 
             "NumberedEquation", FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedEquation", "Presentation"], 
             CellMargins -> {{74, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Condensed"], 
             CellMargins -> {{52, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Printout"], 
             CellMargins -> {{54, 0}, {Inherited, Inherited}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedFigure"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Figure ", 
                   CounterBox["NumberedFigure"]}]], None}}, CounterIncrements -> 
             "NumberedFigure", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedFigure", "Presentation"]], 
            Cell[
             StyleData["NumberedFigure", "Condensed"]], 
            Cell[
             StyleData["NumberedFigure", "Printout"]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedTable"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Table ", 
                   CounterBox["NumberedTable"]}]], None}}, TextAlignment -> 
             Center, CounterIncrements -> "NumberedTable", 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedTable", "Presentation"]], 
            Cell[
             StyleData["NumberedTable", "Condensed"]], 
            Cell[
             StyleData["NumberedTable", "Printout"]]}, Open]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 6],
          
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontWeight -> "Bold"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "\nThe cells below define styles useful for making hypertext \
ButtonBoxes.  The \n\"Hyperlink\" style is for links within the same \
Notebook, or between \nNotebooks.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, 
             FontVariations -> {"Underline" -> True}, FontColor -> 
             RGBColor[0, 0, 1], Background -> None, 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"]], 
            Cell[
             StyleData["Hyperlink", "Condensed"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
         "\nThe following styles are for linking automatically to the on-line \
help \nsystem.\n", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"]], 
            Cell[
             StyleData["MainBookLink", "Condensed"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"]], 
            Cell[
             StyleData["AddOnsLink", "Condensed"]], 
            Cell[
             StyleData["AddOnLink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuideLink", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"]], 
            Cell[
             StyleData["RefGuideLink", "Condensed"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"]], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"]], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]]}, Open]]}, 
     Open]]}, Visible -> False, FrontEndVersion -> 
  "6.0 for Microsoft Windows (32-bit) (April 20, 2007)", StyleDefinitions -> 
  "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{
 "Info3508663653-8109590"->{
  Cell[2710, 121, 120, 2, 58, "Print",
   CellTags->"Info3508663653-8109590"],
  Cell[2833, 125, 851, 24, 45, "Print",
   CellTags->"Info3508663653-8109590"]},
 "Info3508663659-8109590"->{
  Cell[9761, 415, 120, 2, 58, "Print",
   CellTags->"Info3508663659-8109590"],
  Cell[9884, 419, 2535, 67, 192, "Print",
   CellTags->"Info3508663659-8109590"],
  Cell[12801, 511, 122, 2, 58, "Print",
   CellTags->"Info3508663659-8109590"],
  Cell[12926, 515, 2211, 60, 160, "Print",
   CellTags->"Info3508663659-8109590"]},
 "Info3508663660-8109590"->{
  Cell[15482, 596, 122, 2, 58, "Print",
   CellTags->"Info3508663660-8109590"],
  Cell[15607, 600, 849, 24, 47, "Print",
   CellTags->"Info3508663660-8109590"]}
 }
*)
(*CellTagsIndex
CellTagsIndex->{
 {"Info3508663653-8109590", 57262, 1595},
 {"Info3508663659-8109590", 57451, 1600},
 {"Info3508663660-8109590", 57805, 1609}
 }
*)
(*NotebookFileOutline
Notebook[{
Cell[580, 21, 39, 0, 55, "Title"],
Cell[622, 23, 51, 0, 37, "Subtitle"],
Cell[676, 25, 73, 2, 54, "Copyright"],
Cell[CellGroupData[{
Cell[774, 31, 46, 0, 54, "Section"],
Cell[CellGroupData[{
Cell[845, 35, 116, 4, 44, "Subsection"],
Cell[CellGroupData[{
Cell[986, 43, 104, 5, 48, "Input"],
Cell[1093, 50, 448, 10, 23, "Print"]
}, Open  ]],
Cell[1556, 63, 66, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[1647, 67, 101, 2, 48, "Input"],
Cell[1751, 71, 85, 1, 47, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[1885, 78, 43, 0, 44, "Subsection"],
Cell[1931, 80, 66, 0, 48, "Input"]
}, Open  ]],
Cell[CellGroupData[{
Cell[2034, 85, 46, 0, 44, "Subsection"],
Cell[2083, 87, 32, 0, 25, "Text"],
Cell[2118, 89, 97, 1, 48, "Input"],
Cell[2218, 92, 94, 4, 25, "Text"],
Cell[CellGroupData[{
Cell[2337, 100, 23, 0, 48, "Input"],
Cell[2363, 102, 122, 3, 62, "Output"]
}, Open  ]],
Cell[2500, 108, 160, 7, 25, "Text"],
Cell[CellGroupData[{
Cell[2685, 119, 22, 0, 48, "Input"],
Cell[2710, 121, 120, 2, 58, "Print",
 CellTags->"Info3508663653-8109590"],
Cell[2833, 125, 851, 24, 45, "Print",
 CellTags->"Info3508663653-8109590"]
}, Open  ]],
Cell[3699, 152, 112, 3, 25, "Text"],
Cell[CellGroupData[{
Cell[3836, 159, 35, 0, 48, "Input"],
Cell[3874, 161, 328, 10, 62, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[4251, 177, 37, 0, 44, "Subsection"],
Cell[4291, 179, 31, 0, 25, "Text"],
Cell[4325, 181, 59, 4, 82, "Input"],
Cell[4387, 187, 139, 3, 25, "Text"],
Cell[CellGroupData[{
Cell[4551, 194, 25, 0, 48, "Input"],
Cell[4579, 196, 529, 18, 47, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[5157, 220, 37, 0, 44, "Subsection"],
Cell[5197, 222, 27, 0, 25, "Text"],
Cell[5227, 224, 70, 0, 48, "Input"],
Cell[5300, 226, 34, 0, 25, "Text"],
Cell[5337, 228, 165, 7, 133, "Input"],
Cell[CellGroupData[{
Cell[5527, 239, 60, 1, 48, "Input"],
Cell[5590, 242, 320, 9, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[5947, 256, 23, 0, 48, "Input"],
Cell[5973, 258, 502, 15, 67, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[6512, 278, 23, 0, 48, "Input"],
Cell[6538, 280, 499, 15, 67, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[7074, 300, 23, 0, 48, "Input"],
Cell[7100, 302, 497, 15, 67, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7646, 323, 56, 0, 44, "Subsection"],
Cell[CellGroupData[{
Cell[7727, 327, 28, 0, 48, "Input"],
Cell[7758, 329, 120, 1, 23, "Print"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[7927, 336, 53, 0, 44, "Subsection"],
Cell[CellGroupData[{
Cell[8005, 340, 31, 0, 48, "Input"],
Cell[8039, 342, 228, 5, 23, "Print"]
}, Open  ]],
Cell[8282, 350, 33, 0, 25, "Text"],
Cell[8318, 352, 27, 0, 48, "Input"],
Cell[CellGroupData[{
Cell[8370, 356, 24, 0, 48, "Input"],
Cell[8397, 358, 494, 15, 67, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[8928, 378, 24, 0, 48, "Input"],
Cell[8955, 380, 493, 15, 67, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[9497, 401, 120, 5, 44, "Subsection"],
Cell[9620, 408, 66, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[9711, 412, 47, 1, 48, "Input"],
Cell[9761, 415, 120, 2, 58, "Print",
 CellTags->"Info3508663659-8109590"],
Cell[9884, 419, 2535, 67, 192, "Print",
 CellTags->"Info3508663659-8109590"]
}, Open  ]],
Cell[12434, 489, 122, 5, 25, "Text"],
Cell[CellGroupData[{
Cell[12581, 498, 36, 0, 48, "Input"],
Cell[12620, 500, 94, 3, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[12751, 508, 47, 1, 48, "Input"],
Cell[12801, 511, 122, 2, 58, "Print",
 CellTags->"Info3508663659-8109590"],
Cell[12926, 515, 2211, 60, 160, "Print",
 CellTags->"Info3508663659-8109590"]
}, Open  ]],
Cell[15152, 578, 34, 0, 25, "Text"],
Cell[15189, 580, 79, 2, 48, "Input"],
Cell[15271, 584, 136, 5, 25, "Text"],
Cell[CellGroupData[{
Cell[15432, 593, 47, 1, 48, "Input"],
Cell[15482, 596, 122, 2, 58, "Print",
 CellTags->"Info3508663660-8109590"],
Cell[15607, 600, 849, 24, 47, "Print",
 CellTags->"Info3508663660-8109590"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[16505, 630, 52, 0, 44, "Subsection"],
Cell[CellGroupData[{
Cell[16582, 634, 78, 1, 48, "Input"],
Cell[16663, 637, 1301, 59, 919, "Print"]
}, Open  ]],
Cell[17979, 699, 163, 5, 55, "Text"],
Cell[CellGroupData[{
Cell[18167, 708, 76, 1, 48, "Input"],
Cell[18246, 711, 111, 2, 47, "Output"]
}, Open  ]],
Cell[18372, 716, 113, 3, 25, "Text"],
Cell[CellGroupData[{
Cell[18510, 723, 170, 8, 150, "Input"],
Cell[18683, 733, 122, 1, 23, "Print"]
}, Open  ]],
Cell[18820, 737, 57, 0, 48, "Input"],
Cell[CellGroupData[{
Cell[18902, 741, 42, 0, 48, "Input"],
Cell[18947, 743, 111, 2, 47, "Output"]
}, Open  ]],
Cell[19073, 748, 46, 0, 25, "Text"],
Cell[CellGroupData[{
Cell[19144, 752, 368, 9, 69, "Input"],
Cell[19515, 763, 72, 1, 47, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[19624, 769, 72, 1, 48, "Input"],
Cell[19699, 772, 156, 5, 55, "Print"]
}, Open  ]],
Cell[19870, 780, 35, 0, 25, "Text"],
Cell[19908, 782, 73, 0, 48, "Input"]
}, Open  ]]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
