(* Content-type: application/vnd.wolfram.mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 8.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       157,          7]
NotebookDataLength[    129588,       3200]
NotebookOptionsPosition[     82164,       2087]
NotebookOutlinePosition[    119574,       2903]
CellTagsIndexPosition[    119531,       2900]
WindowFrame->Normal
ContainsDynamic->False*)

(* Beginning of Notebook Content *)
Notebook[{
Cell["MathCode Introduction", "Title"],

Cell["Copyright \[Copyright] MathCore AB 1999", "Copyright"],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell["\<\
MathCode is an application package which includes the following:\
\>", "Text"],

Cell["\<\
Translation of a subset of the Mathematica language with type information to \
efficient C++ code\
\>", "Itemize"],

Cell["Type annotations compatible with standard Mathematica", "Itemize"],

Cell["\<\
Availability of Matlab-like matrix operations on array sections both in \
Mathematica and in compiled code\
\>", "Itemize"],

Cell["\<\
Transparent calling of compiled executable code via MathLink or stand-alone \
execution\
\>", "Itemize"],

Cell["\<\
The performance of compiled generated C++ code is often approximately a \
factor of 1000 better than standard interpreted Mathematica, and often a \
factor of 100 better than code compiled using the internal Mathematica \
compiler.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["The MathCode System", "Section"],

Cell["\<\
The MathCode application can easily be used just by executing the following \
command in Mathematica:\
\>", "Text"],

Cell["Needs[\"MathCode`\"];", "Input",
 InitializationCell->True],

Cell[TextData[{
 "If you are compiling your own package, e.g. called ",
 StyleBox["foo",
  FontFamily->"Courier New"],
 ", using MathCode, you also need to mention MathCodeContexts within the path \
of the package as below."
}], "Text"],

Cell[BoxData[
 RowBox[{"BeginPackage", "[", 
  RowBox[{"\"\<foo`\>\"", ",", 
   RowBox[{"{", "MathCodeContexts", "}"}]}], "]"}]], "DisplayCode"],

Cell[CellGroupData[{

Cell["Code Generating Phase", "Subsection"],

Cell[TextData[{
 "The MathCode code generator translates a Mathematica package, here called \
foo, to a corresponding C++ source file, here called ",
 StyleBox["foo.cc",
  FontFamily->"Courier"],
 ". Additional files which are automatically produced are: the header file ",
 StyleBox["foo.h",
  FontFamily->"Courier"],
 ", the Math header file ",
 StyleBox["foo.mh",
  FontFamily->"Courier"],
 " which stores MathCode information, the MathLink related files ",
 StyleBox["footm.c",
  FontFamily->"Courier"],
 ", ",
 StyleBox["foo.tm",
  FontFamily->"Courier"],
 ", and ",
 StyleBox["fooif.cc",
  FontFamily->"Courier"],
 " which enable transparent calling of the C++ versions of functions  in foo \
from Mathematica, and foomain.cc which contains the function main needed when \
building a stand-alone executable for foo."
}], "Text",
 CellChangeTimes->{{3.5086611376496224`*^9, 3.508661139024605*^9}}],

Cell[GraphicsData["Metafile", "\<\
CF5dJ6E]HGAYHf4PEfU^I6mgLb15CDHPAVmbKF5d0@0002A@0@0005P0000000007@000:H2002A0@00
000000000005A`00=BL00215CDH00040D2@009/000080000000000000000000080<005P2003C0000
WP0004H000342@00^0T004M4BD<1002000<000Oj^f/00000X0T000402@000lL40007054000000140
000V1Pl0603ooooo000@083hoolSnOooV`P000L2000900009PH?00P0ooooo`80000>00009PH?0180
ooooo`@02P1DCU1@0`0hDGd:2P0002H63`0:05A>D5000080l0<900009PH?00P0ooooo`<0000?0000
9PH?01@0E4i@D0@0300100000@000000000500002`8SnH3h1@0000`2i0PK40T0003j0PD000000?oo
o`0R00@0000]0@001`000?`20@00000000040000;@4100T0003j0PH020000000008R00@0000]0@80
1`000?`2000c<l`200040000;@4300D000090S<cc0840000104=00L0000K157n`Og@nH3h5@000?/2
@?l00000002@0Ol00000000BE6U]IG<PCVEg859_KF5^0:h2100002d110040000;P4H00D0000:0P00
000500002@;oool21@0001@2000000@000020@406000038:4?`@n@/0001=HGAXIFeQM6USH@2P0600
=@1P05H0RP1P03H0=@1E0600100000810P0500002@8c<l`21`0001/4TOgQ1`3jd08E0000n`90o`00
00000901o`000000019DJFeULb1>IGLPDVm]HFh06@040000;@4500@0003`0@@0100002h160050000
2P8000001@0000T2oooo0PD0000D0_oo3I<400000P4100d0000b2[3kl0<40000CF5dJ:00H00e0600
100000810P0E0000n`90o`000000090100000000019DJFeULb1>IGLPDVm]HFh0[P840000;@4400@0
003`0@D01@0000T2oooo0PD0000D0_oo3I<400000P4100d0000b2[3kQ@D40000@fmTIH00H01P05D0
100000810P0E0000n`90o`000000090100000000019DJFeULb1>IGLPDVm]HFh06@040000;@4500@0
003`0@@0100002h1600500002P8000001@0000T2oooo0PD0000D0_oo3I<400000P4101D0000b2YKl
l0<90000AfE^IG9QM6mb08/0E@1P05D0@01E03H0H01000@000020@801@0000T200000P@0000]0@00
100002d10@040000l04200@0003`0@<040000?/240070000002l0P0000010P8RDgUcM6E]06h40000
;@4200@0003`0@D00`0001h01`0001H41T1l0P70`?d90000nP8600P0000000028P040000;@4300D0
000D0R3k__d500004`8PngH2100002d100040000;@4100@0003`0@<0100002L1ool70000o0800000
008000@0000]0@<01@0000T200000P`0000T0`@0EP96nfD28?]F0_[jd08PnaD0003k0Uoo00000000
T040000000004UAYKFEc84iUMb1BKfeQKP2^0P@0000]0@@0100002h1600500002P8000001@0000T2
00000PD0000D0R3kMP8400000P4101P0000b2]3iD?h;0000CF5dJ6E]HGAYHf40SP1702d0D01707`0
A`0]02`0A`1700@000020@805@000?/2Gol00000002@0@000000000BE6U]IG<PCVEg859_KF5^01T0
100002d11@040000l04400@0000^0AP01@0000X2000000D000090P000085000050;oo`fC10000081
0@0L0000<PZ@nU3n3P00071QHf]QIfEc;fEhL78^D01704L0D0170500A`0n02d0A`1@0500=@0X00@0
00020@801@0000T200000P@0000]0@00100002d10@040000l04300@0000]0@8010000?011@030000
7P0700005P@6@=P20L0KoPT0003j0PH020000000008R00@0000]0@<01@0001@2@?`IoPD0000C0T3l
dP840000;@4000@0000]0@4010000?010`0400009`7oo`L0003l0P0000000P00100002d10`050000
2@800002300002@3100joQ[l:oi0o3[nI_c0oD3l5@000?/2Gol00000002@0@000000000BE6U]IG<P
CVEg859_KF5^0:h2100002d110040000;P4H00D0000:0P00000500002@8000021@0001@2@?cB0P@0
00020@404`00038:<?gPoPP0001CNFeRKfaYHeT0D01l0500D00]02`0A`0400000P4201D0003k0Uoo
00000000T040000000004UAYKFEc84iUMb1BKfeQKP0I00@0000]0@D010000?0110040000;P4H00D0
000:0P00000500002@8000021@0001@2ool=T`@000020@405P00038:l?gPoPX00015MV5/MF5dJFm^
HP1@04L0;01@04L0;@0/0500D00400000P4200T0003j0PH020000000008R00@0000]0@@01`000?`2
0000c9T200040000;@4600@0003`0@<01@0000T20<bI0Q80000T0`L0n@@]0<d54OmS1A7oH`GdoHh4
m?f>117o90@AoaD0003k0T3o00000000T040000000004UAYKFEc84iUMb1BKfeQKP2^0P@0000]0@<0
10000?011@040000;P4H00D0000:0P00000500002@8000021@0001@2ool=T`@000020@40D@00038:
/041nS40001VKfl^Hb`PIVm_;VP/86I_KfeQJFh^Hb`PIVm_KF`^Hb`PIVm_;WA];21VKfl^HV5d0400
H01P0300E@0`0300@01P0600<01P0300<0100600H02E05H0=@1P0300E@0`0300@01P0600U@0f0300
E@0`0300@01P0600<00e09H0<00`0400H01P0300H01E03D0100000810P0?00009PH?01@0E4i@D0@0
3000000000000000000900009PH?00P0ooooo`4000040000;@4000@0000]0@4010000?0110040000
l04600@0000]0@8010000?010`030000000000000000000000000000000A0000300000P0000;0000
40000;42001m0@002@000100002a0P00O@4000X0000@0000P?Soob?iool90000400001/@003T2000
600000`0000c<l`29P0001`0000100001P0000P00000000000000RD0000<00000@0002L0000H0000
0P000000000c<l`2000002D0000<00000P0002/0000H0000P?Soom3ioonloOooBokooaH0000<0000
600001P0000<0000oooo0Q80000<00000@0001/0000@000000000000001B0000C04000<00010oooo
0000000000000000T0400?l00000000BE01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h0000`0000
000009RK]P000000000000000000000000000000000B000000000900000000000000000000000000
00000642000C00000000000000000000000000000000000000000682000a000000000030NL@0`7W4
0000000000000000000006H20002000000000000000000000000000000000000000006L200020000
00000000000000000000000000000000000006P2000@000000000000000000000000000000000000
000000410@410@410@4002D0000<00000`0005@0002@0000600007d000300000X@000040002n2Xe0
hdR=@13iool@o?oo2`0004`0000000000000000000000000000006@0001=0640M01X06D0K@1Q07@0
J@1S0640002Q0000H0000300001V0000D`0008X0001P0000=P0003D0001D0000H0000180000<0000
0P0001P0000<0000<c?<0R/0000H0000c`80003joooL1`00S?gooe80001<0@001000043oool00000
00000000002@0@00o`000000019D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP00074200020000
00000040000000000000000000000000000007820006000000000030NL@000000000000000000000
000007<200040000000000L0000000000000000000000000000007@2000e000000000030NL@0`7W4
0<1ia030NL@00000000007T20002000000000040000000000000000000000000000007X2000F0000
00000000P3l0000000000000000000000@0000000000000000000000000000000000000000000000
0@410@410@410@009@0000`000040000:00000`000030000600000`0003oool24P0000`000010000
6`000100000=ToooooooobD0000<0000100005@0001T0000j@4006d0000/0P00T@000040002n2Xe0
hdR=@?03002`nooo100004`0000000000000000000000000000005@0001=0640M01X0:80001O0000
=P000600000B000030000080001B0000C04000<00010oooo0000000000000000T04000000000000B
E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000Y`400080000000000000000000000000
0000000000000000Z04000800000000000000000000000000000000000000000Z@40008000000000
0@000000000000000000000000000000ZP4000`000000000Q5MJ0000000000000000000000000000
Z`4000800000000000000000000000000000000000000000[@400080000000000000000000000000
0000000000000000[P400000000000000000000000000000000000410@410@410@4002D0000<0000
0`0002P0000<000010000180000<00000@0001/0000@00003I?ooooooolU0000300000<0001D0000
I00002d2001]0000L080090000010000_PZ=@>=8SD251@00/?_oo`@0001<00000000000000000000
00000000001D0000@`1_06@0I@240000G`000600001D00004P0000`000020000DP0004`100040000
@?ooo`0000000000000009010000000000004U@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^0000
^04002`000000000oh;ooon2ool000000000000000000000^`4003P000000000[4Nf08d_]P1/B;H0
K4Rf08d_]P2/AkH0`P4003P000000000SBnf08d_]P2=;kH0SBnf08d_]P2=;kH0b@4002h000000000
0@400000000000000000000000000000c04002h00000000000000000000000000000000000000000
d04003@000000000P0@0060300201000H0<0000000000000e@4002`0000000000000000000000000
0000000000010@410@410@41000U0000300000@0000X0000300000<0000B000030000040000K0000
400000fCoooooooo9@0000`000040000E00008@0003Y0@00U00006P2002g00000@000;h:SD3SB8e0
l0<009Klool90000C0000000000000000000000000000000H00004L0I@1^06D0LP1Q07@0K`1b0000
RP0005<0001P0000E0000480001D0000=@000600001200004P0000`000020000600000`000000002
9P0001`0000300001@00000000000000oooo02D0000<00000`0002P0000<00000@0002L0000H0000
0@00004000000000000002D0000<00000@0002P0000<00000P000580001<0@000P00010000070000
00000000002l0P00000000420R9C07T0L`1d06D0K@0000000000000000000000h`4000@000000000
oh;oo`00000000000000000000000000000000@000000000oh;oo`00000000000000000000000000
iP4003`00000000000000800000000000000000000000000i`400080000000000000000000000000
0000000000000000j04000800000000000000000000000000000000000000000j@40008000000000
0@000000000000000000000000000000j`4000D00000000000000000000000000000000000000000
0@410@410@410@009@0000`000020000:00000`0000400009@0000`0000300009@0000`000010000
9@0000`0000200008@0000P0000N000060000<3mool1`?ooO08000I0000K000040000;kmoolPnooo
9P0001`0000400001P0000P00000000000000RD0000<0000100003H0000@0000MP80023koolU0000
300000<0000X0000300000@0000U0000300000<0000U000030000040000R000030000?oooolW0000
600000@00000000000000P00000U0000300000@0001F0000;0000:D1001?0000^@4005`000040000
EP96nfD28?]F0_[jd08Pna80000<00000@0001/0000@0000MP80023koomB0000C04000D0001Ooooo
0000000000000000T04000000000000BE01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h000000000
0000000000000000nP4003@000000000P000000000200000X0000000000000000080018000000000
100000000000000000000000000000000@8001`00000000000000?oo000000000000000000000000
0P8002`00000000000000200000000000000000000000000000002`0000000000000070000000000
00000000000000001@80018000000000`00000000000000000000000000000000000000000000000
000000410@410@410@4002D0000<00001@0005@0002@0000n@0001d000230@00>`000040002n2Xe0
hdR=@53nooo@nOoo2`0004`0000000000000000000000000000006@0001=0640M01X06D0K@1Q07@0
J@1S0640002:0000B0000300001=0000B00007h000180000:@0002X000180000B0000180000<0000
0P000580001<0@001P0005ooool0000000000000002@0@0000000000019D06T0K@1U07<0801>06D0
M`0P0580K`1]0640KP0000000000000000000000000000004P800080000000000000000000000000
00000000000000004`8003d00000000010000?ooool0000000000000000000005080048000000000
SBnf00000000000000000000000000005@80048000000000SBnf0000000000000000000000000000
5P800080000000000@0000000000000000000000000000005`8000@0000000001000000000000000
000000000000000000000000000000000@410@410@410@009@0000`000060000:00000`000050000
4P0000`0000100006`000100000=ToooooooobD0000<00001P0005@0002P0000n@0003d0002G0@00
F`000040002n2Xe0hdR=@53noon@n_oo3P0004`0000000000000000000000000000006P0001`0640
H`1[0640I`1U07<0;`1U07P0L01b02h0C@0004P000180000CP0004P0001=0000B00003`0000`0000
B00004d0001D0000<00002X0000B000030000080000U000030000040000X0000300000@0000U0000
30000080000X0000300000H0000U0000300000<0000U000030000040000U000030000080000Q0000
200001h0000H00006okoo`70oooH0P001T0001/0000@00006Okood3loolV0000700000H000060000
20000000000000029@0000`000060000=P000100003B0P00@?coobD0000<00000`0002P0000<0000
1P0002D0000<00000`0002D0000<00000@000280000<0000ooooobL0000H00001P00000000000002
000002D0000<00001P0005H0000/0000h@0007l0003e0000S00000@0000joQ[l:oi0o3[nI_c0oD3l
4P0000`0000100006`000100003B0P00@?cooe80001<0@00100005ooool0000000000000002@0@00
00000000019D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP0000@0000000003H7oo`0000000000
0000000000000000000000`000000000F6MJ0000000000000000000000000000=`8000`000000000
F6MJ0000000000000000000000000000>08000d000000000H7IJ0000000000000000000000000000
>@8000d00000000026MJ0000000000000000000000000000>`8000d000000000C6MJ000000000000
0000000000000000?@8000d000000000000000000000000000000000000000000@410@410@410@00
9@0000`000040000E00007`0000A0@00[P0007H1003<00000@000;h:SD3SB8e0h?kooc3mool80000
C0000000000000000000000000000000G00005<0N@1]0680K`1/06T0H`1I0000CP0007h0001>0000
C@000300000Z0000B0000180000<00000P000580001<0@001@0005ooool0000000000000002@0@00
00000000019D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP000000000000000000000000000000
BP8000800000000000000000000000000000000000000000B`800?oooon8TKP00000000000000000
0000000000090`00<0000?oooon@TKP0000000000000000000000000000:0`00<0000?oooonXTKP0
00000000000000000000000000000000:@000?oooomhTKP0000000000000000000000000000<0`00
80000?oooon0TKP00000000000000000000000000000000000000000000000000@410@410@410@00
9@0000`000050000:00000`0000400004P0000`0000100006`000100000=ToooooooobD0000<0000
1@0005@0002800004@400<h000250@00k0000040002n2Xe0hdR=@>3nooo`oOoo2P0004`000000000
000000000000000000000600001507H0H@1/07D0H@1d06T0K`1^06H0001=0000B00002X0001D0000
B00002X0000Y0000E00004h0000B000030000080000W0000600000@0000000000<bI0P00000U0000
300000@0000X0000300000H0000H00003000003<V@8V0000700000H0000600002000000000000002
9@0000`0000600009@0000`000040000EP0003P0003b0@00cP0003T2000^0@001`000?T4;@3=1A7o
H`DAof<5m?f>1?CmSP@Aob@44OmB0000C04000L00010oooo0000000000000000T04000000000000B
E01Y06d0I@1c0200CP1U07L0801B06l0K@1Q06h00000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000000
000000000000000000000000000000000000000000000000000000410@410@410@4002D0000<0000
1`0002P0000<00001@0001P0000<000000000Q80000<00000@0001/0000@00003I?ooooooolU0000
300000L0001D0000M0400400001^0@00YP80094100010000_PZ=@>=8SD01n_oo/0400340001<0000
000000000000000000000000002`0000IP1_06l0;P1S02`0801V06l0K`0^06P0;00P06H0K`1_06d0
H@1Y06h0;P1S02`0801V06l0K`1]06`0;P1S02`0801V06l0K`0^07@0K@0/0200IP1_06l0;P1R0640
M0000480001O0000H0000300001D0000<000030000110000H0000600000`0000G`000300000`0000
@P000600001P0000U@0005@0000f0000G`000300001D0000<000030000120000G`000600002F0000
=P000300001C0000<000030000120000H00005l0000`0000=P0009H0000`0000;`000480001P0000
H0000300001O0000E00003H0000B000030000080000U0000300000<0000X0000300000H0000U0000
30000040000X0000300000@0000U000030000080000X0000300000L0001;00004000000000050000
9@0000`000070020:00000`0000300009@0000`000000020:00000`0000100009@0000`0000=0020
:00000`0000200003P0001@000000000400001@0
\>"], "NumberedFigure",
 ImageSize->{391.188, 214.563},
 ImageMargins->{{0, 0}, {0, 11}},
 ImageRegion->{{0, 1}, {0, 1}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Building Phase", "Subsection"],

Cell[TextData[{
 "The generated file ",
 StyleBox["foo.cc",
  FontFamily->"Courier"],
 " created from the package foo together with header file ",
 StyleBox["foo.h",
  FontFamily->"Courier"],
 " and additional files are compiled and linked into two executables. \
Numerical libraries may be included in the linking process."
}], "Text"],

Cell[GraphicsData["Metafile", "\<\
CF5dJ6E]HGAYHf4PEfU^I6mgLb15CDHPAVmbKF5d0@0002lL0@0005P0000:0000000000/3003d0000
00000000001PD000dQP00215CDH0004072l001D100060000000000000000000080<005P2003C0000
WP0004H0000T3P0060h004M4BD<1002000<00?ADUFh0000000h000402@000o86000601P000000140
000V1Pl0603ooooo000@0<Ofoom@nooo0@T00?40000900009PH?00P0ooooo`80000>00009PH?0180
ooooo`@02P1DCU1@0`0hDGgo2P0002H63`0:05A>D5000080l0<900009PH?00P0ooooo`<0000?0000
9PH?01@0E4i@D0@0300100000@000000000500002`9@nlOf1@0000`2X@Dj4PT0003j0PD000000?oo
o`0R00@0000]0@001`000?`20@00000000040000;@4101D0003k0T3o00000000T040000000004UAY
KFEc84iUMb1BKfeQKP0]00@0000]0@80100002h1600500002P8000001@0000T200000PD0000D0P00
000400000P4100l0000b2T3l0OL50000IVm_;V<0@01P0600<01E00@000020@805@000?/2@?l00000
002@0@000000000BE6U]IG<PCVEg859_KF5^02d0100002d10`040000l04200@0000^0AP01@0000X2
000000D000090P000085000050;oojfB100000810@0?0000<PXVo@7g1@0006I_KbiX0400H01P0300
H00400000P4201D0003k0T3o00000000T040000000004UAYKFEc84iUMb1BKfeQKP0]00@0000]0@80
10000?010`040000;P4H00D0000:0P00000500002@8000021@0001@2oon]TP@000020@405@00038:
3?h1m`T0001VKfm]HFU^;V<0@01P0600U@1F03D0H00`05D0100000810P0E0000n`90o`0000000901
00000000019DJFeULb1>IGLPDVm]HFh0;@040000;@4300@0003`0@80100002h1600500002P800000
1@0000T200000PD0000D0_oo[I8400000P410180000b2_;n0OL70000IVm_KF`^H`100600H02E03H0
<01E00@000020@805@000?/2@?l00000002@0@000000000BE6U]IG<PCVEg859_KF5^02d0100002d1
0P040000l04300@0000^0AP01@0000X2000000D000090P000085000050;oojfB100000810@0@0000
<P[Ho`7g1P0006I_KbidKD00H01P0300=@2F00@000020@802@000?X21P08000000000R80100002d1
0`070000o080003<V@8000@0000]0@@01@0000T20<bI0P@000040@d01`0001/4TOf10E3kX?`E0000
n`90o`000000090100000000019DJFeULb1>IGLPDVm]HFh0;@040000;@4500@0003`0@80100002h1
600500002P8000001@0000T200000PD0000D0_oo[I8400000P4101P0000b2]3l<?d;0000IVm_KF5Y
KRiUN6D0@01P0600U@1F03D0H00`05D0H01E00@000020@801@0000T20<bI0PL0000K1?40P@6`oZ3l
5@000?/2@?l00000002@0@000000000BE6U]IG<PCVEg859_KF5^02d0100002d10P040000l04500@0
000^0AP01@0000X2000000D000090P000085000050;oojfB100000810@0E0000<PX`033m2@0006I_
Kfe/;VEhI@100600H02E03H0<01E0600E@0400000P4200L0003l0P00UYJF0P00100002d11@040000
l04400D000090YJFUP8700006`AQ0049D?]@11D0003k0T3o00000000T040000000004UAYKFEc84iU
Mb1BKfeQKP0]00@0000]0@@010000?010P040000;P4H00D0000:0P00000500002@8000021@0001@2
oon]TP@000020@405@00038:`?d@1@T0001>MFeULVUSHF`0R`1P09D0E@1003H0E@1E03D010000081
0P0E0000n`90o`000000090100000000019DJFeULb1>IGLPDVm]HFh0;@040000;@4200@0003`0@@0
100002h1600500002P8000001@0000T200000PD0000D0_oo[I8400000P410180000b2ZKn40D70000
L65SJf5WI@1P05D0E@1P05H0H01E00@000020@801@0000T200000P@0000]0@00100002d10@040000
l04300@0003`0@D040000?/240070000002l0P0000010P8RDgUcM6E]06h40000;@4300@0003`0@80
0`0001h01`0001H41T3/n`70X?T90000nP8600P0000000028P040000;@4200D0000D0Q3lX?T50000
4`8no>Ck100002d100040000;@4100@0003`0@80100002L1ool70000o0800000008000@0000]0@80
1@0000T200000P`0000T0`@0`O]Ro=?k?Oc8naKl?_a6o0D000090P0000840000;@4000@0000]0@40
10000?010P0300007P0700005P@6@1gl0L2Pn@T0003j0PH020000000008R00@0000]0@801@0001@2
d?bNn@D0000C0]3l5o`40000;@4000@0000]0@4010000?010P0400009`7oo`L0003l0P0000000P00
100002d10P0500002@800002300002@3103fnoKl1Oc@o?KkZ_a`o=3l1@0000T200000P@0000]0@00
100002d10@040000l04200<0000N00L0000F10I0k_/1`33j2@000?X21P08000000000R80100002d1
0P05000050;0oC3j1@0001<2L?gVn`@0000]0@00100002d10@040000l04200@0000W0Ooo1`000?`2
0000000200040000;@4200D000090P00008<000090<40<okVogGng?m`O]AoD3lH?d500002@800002
100002d100040000;@4100@0003`0@800`0001h01`0001H4Noh6o13lL?T90000nP8600P000000002
8P040000;@4200D0000D0Q3lL?T500004`9co_kk100002d100040000;@4100@0003`0@80100002L1
ool70000o0800000008000@0000]0@801@0000T200000P`0000T0`@0cO]io_;kIoh1o47n@?b`oPD0
00090P0000840000;@4000@0000]0@4010000?010P0300007P0700005P@6@<?k0L00nPT0003j0PH0
20000000008R00@0000]0@801@0001@2h?h0nPD0000C0X7o^o/40000;@4000@0000]0@4010000?01
0P0400009`7oo`L0003l0P0000000P00100002d10P0500002@800002300002@3102@ni[oZo]loj[k
Dol@o:3o1@0000T200000P@0000]0@00100002d10@040000l04200<0000N00L0000F10_oe_/0oJ3i
2@000?X21P08000000000R80100002d10P0500005080oJ3i1@0001<20oo>n`@0000]0@00100002d1
0@040000l04200@0000W0Ooo1`000?`20000000200040000;@4200D000090P00008<000090<409gk
2Oo1noSnd?_AoQ3l@?l500002@800002100002d100040000;@4100@0003`0@800`0001h01`0001H4
1T3cn`70X?T90000nP8600P0000000028P040000;@4200D0000D0Z3oX?T500004`9b0>_k100002d1
00040000;@4100@0003`0@80100002L1ool70000o0800000008000@0000]0@801@0000T200000P`0
000T0`@0`?^;0=_kK03Jnd<0@?b@00D000090P0000840000;@4000@0000]0@4010000?010P030000
7P0700005P@6@?P30L0k0PT0003j0PH020000000008R00@0000]0@801@0001@2L?cb0`D0000C0W3l
>@840000;@4000@0000]0@4010000?010P0400009`7oo`L0003l0P0000000P00100002d10P050000
2@800002300002@3101J0T[lB`9`o5X2U_cP0G3l1@0000T200000P@0000]0@00100002d10@040000
l04200<0000N00L0000F10I0n0<1`3/22@000?X21P08000000000R80100002d10P05000050:Poo83
1@0001<2X?li0P@0000]0@00100002d10@040000l04200@0000W0Ooo1`000?`20000000200040000
;@4200D000090P00008<000090<405X2N_m;0Z3oFP;6on01X?l?00009PH?01@0E4i@D0@030000000
00000000000900009PH?00P0ooooo`4000040000;@4000@0000]0@4010000?010P03000000000000
000000000000000000000000000000004@0000`0000800002`000100000<0`00l@0000T0000@0000
30<00?40000:000040000<Ofoom@nooo2@000100000j4P00X@D001H0000<0000600001P0000<0000
00000Q80000<00000@0001/0000@000000000000000W0000600000400001000000000000000U0000
30000040001B0000C04000800010oooo0000000000000000T04000000000000BE01Y06d0I@1c0200
CP1U07L0801B06l0K@1Q06h0003oP_oooh;ooon2ool00000000006T1000d00003`000000ooooP_oo
@0T00<0>00000000000006/1000d000000000?n2ooooP_oooh;ooon2ool00000000006d1000/0000
0`000;0/003`70000000000000000000000006h1000d00003`000?n2ooooP_oo/000050000000000
00000701000d00003`000?n2ooooP_ooH0T0060;0000000000000781000<000000000<`]^P000000
00000000000000000000000000000000000000410@410@410@4002D0000<00000P0005@0001/0000
2P0002P0001:0000B`000040002g5Xe0WRN=@07goom0o?oo1@0004`0000000000000000000000000
000005P0001V06l0K`0^06<000120000H00005l0000`0000E0000180000<00000P000580001<0@00
0`00043oool0000000000000002@0@0000000000019D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP000040000000000000000000000000000007`1000n0000000000T0000000000000000000000000
000007d1000n0000000000<0000000000000000000000000000007h1000n0000000000D000000000
0000000000000000000007l100020000000000000000000000000000000000000000080100020000
00000000000000000000000000000000000008410002000000000040000000000000000000000000
00000000000000000@410@410@410@009@0000`000030000:00000`0000200004P0000`000010000
6`000100002]T_ooooooobD0000<00000`0005@0001/00002P0004l0001<0000LP000040002g5Xe0
WRN=@07goolVoOoo1@0004`0000000000000000000000000000005P0001V06l0K`0^06P000120000
H00005l0000`0000H0000180000<00000P000580001<0@000P00043oool0000000000000002@0@00
00000000019D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP00000000000000000008X100020000
00000000000000000000000000000000000008/1000n000000000000000000000000000000000000
000008`10005000000000040000000000000000000000000000008l1000<0000000007AWFP000000
00000000000000000000090100000000000008AGFP000000000000000000000000000941000n0000
00000000000000000000000000000000000000000000000000000000000000000@410@410@410@00
9@0000`000020000:00000`0000300004P0000`0000100006`000100002]T_ooooooobD0000<0000
0P0005@0002400002P0007D0002:0000V0000040002g5Xe0WRN=@07gool<o_oo2@0004`000000000
000000000000000000000600001V06l0K`1]0640J@1^02h0H`000480001P0000H00009D0001D0000
=P0005l0000`0000E0000180000<00000P000580001<0@000`00043oool0000000000000002@0@00
00000000019D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP00000000000000000009h1000d0000
00000?n2ooooPP00003ooon2ool0000000000:<10005000000000040000000000000000000000000
00000:@1000n0000000003l000000000000000000000000000000:D1000n0000000000L000000000
000000000000000000000:H1000000000000004000000000000000000000000000000:L100020000
00000000000000000000000000000000000000000000000000000000000000000@410@410@410@00
9@0000`000030000:00000`0000200004P0000`0000100006`000100002]T_ooooooobD0000<0000
0`0005@0001h00002P0009`0001/0000_`000040002g5Xe0WRN=@07gooobo_oo1`0004`000000000
0000000000000000000005`0001V06l0K`1]06`0;P1S0000@P0005l0001P0000UP0003D0000`0000
E0000180000<00000P000580001<0@000P00043oool0000000000000002@0@0000000000019D06T0
K@1U07<0801>06D0M`0P0580K`1]0640KP00002m00000000000000000000000000000;81000B0000
0000000000000000000000000000000000000;<1000B000000000000000000000000000000000000
00000;H1000@00000000004000000000000000000000000000000;L1000@00000000004000000000
000000000000000000000;P1000/000000000?n2ooooP_oo000000000000000000000;/1000h0000
00000:a7]P2=;kH0000000000000000000000000000000000@410@410@410KH09@0000`000020000
:00000`0000300004P0000`0000100006`000100002]T_ooooooobD0000<00000P0005@0001`0000
2P000<80001M0000i@000040002g5Xe0WRN=@07goooHoooo1P0004`0000000000000000000000000
000005P0001V06l0K`0^07@0K@0k0000H0000600000`0000=@0009H0000B000030000080000H0000
3000003<V@8V0000700000<00006000020000000000000029@0000`0000300009`0001P000040000
0000003<V@8000009@0000`000040000:`0001P0002Oo?ooD?_oog/1002>oOooDP0004`100050000
@?ooo`0000000000000009010000000000004U@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^0000
00000=h1000200000000004000000000000000000000000000000=l1000200000000000000000000
000000000000000000000>01000200000000000000000000000000000000000000000>4100050000
0000004000000000000000000000000000000>81000F000000000000P3l000000000000000000000
00000><10004000000000?n2ool00000000000000000000000000>@1000000000000000000000000
0000000000010@410@410@41000U0000300000D0000X000030000080000H000030000000008B0000
30000040000K000040000:fBoooooooo9@0000`000050000E0000900000B0@00@0000;01001S0000
0@000;LFSD2N9he0<?goom3lool;0000C0000000000000000000000000000000I00006H0K`1_06d0
H@1Y06h0;P1U07P0I@000480001O0000H00009H0001C0000=P000600000`0000E00005l0001D0000
4P0000`000020000600000`00000c9T2:`0001P0002Oo?oo[Okoog/1003[0000DP0004`100020000
@?ooo`0000000000000009010000000000004U@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^0000
000000000000000000000?41000<000000000<AV00000000000000000000000000000?81000j0000
0000000000000000000000000000000000000?<1000<00000000091GFP0000000000000000000000
00000?H1000200000000004000000000000000000000000000000000000200000000000000000000
000000000000000000000?P100020000000000000000000000000000000000000000000000000000
0000000000010@410@410@41000U000030000080000X0000300000D0000H000030000000008B0000
30000040000K000040000:fBoooooooo9@0000`000020000E00008@0000B0@00d@000981003d0000
0@000;LFSD2N9he0<?gooc0000090000C0000000000000000000000000000000H00006H0K`1_06d0
K00^06D0N01U0000@P000600001P0000U@0003H0000`0000E00005l0001D00004P0000`000020000
9`0001P000050000000009JFUP8000009@0000`000050000:00000`000040000600000`0002FUYH2
9@0000`000050000:`0001P0001?1000D?_ooo/8001L0000DP0004`100040000@?ooo`0000000000
000009010000000000004U@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^00000000000000000000
000000`2000200000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000010@41
0@410@41000U0000300000@0000X000030000080000H000030000000008B000030000040000K0000
40000:fBoooooooo9@0000`000040000E00008@0001S0P00J0000>T2002;00000@000;LFSD2N9he0
40D00<3mool90000C0000000000000000000000000000000H00004h0M@1]06D0LP1Y06<0H@1/05D1
RP000600002E0000E0000480000f0000D`0005@0000f00004P0000`000020000DP0004`100020000
@?ooo`0000000000000009010000000000004U@0J@1]06D0L`0P04h0I@1g0200DP1_06d0H@1^0000
T9^f0000000000000000000000000000k@0002400000000049nf0000000000000000000000000000
kP00024000000000<:2f0000000000000000000000000000k`0000h000000000B6mJ000000000000
0000000000000000l00000h000000000?6mJ0000000000000000000000000000l@00024000000000
B9nf0000000000000000000000000000lP00024000000000D:6f0000000000000000000000000000
0000000000010@410@410@41000U000030000080000X0000300000@0000B000030000040000K0000
40000:fBoooooooo9@0000`000020000E00007P0001S0P00S`000</2002b00000@000;LFSD2N9he0
40D00:Knool70000C0000000000000000000000000000000G0000700H@1S06/0H@1W06D0001P0000
D`0005X0001P0000E00005l0001D00004P0000`0000200009P0001`0000400001@00000000000000
oooo02D0000<0000100002P0000<00000`0002D0000<00000@0002P0000<00001@000580001<0@00
1@0001000007000000000000002l0P00000000420R9C07T0L`1d06D0K@0000000000000000000000
o`000080000000000000000000000000000000000000000000000080000000000000000000000000
00000000000000000P400080000000000@0000000000000000000000000000000`40010000000000
200000000000000000000000000000001P4000D0000000000P000000000000000000000000000000
1`4000@000000000l=Soo`0000000000000000000000000020400080000000000000000000000000
00000000000000000@410@410@410@009@0000`000050000:00000`0000200009@0000`000040000
9@0000`0000100009@0000`0000500008@0000P0000N000060000:3iool1`?ook?_oo`I0000K0000
40000:3iool@o?oo9P0001`0000200001P0000P00000000000000RD0000<00000P0003H0000@0000
i?_oockloolU0000300000@0000X000030000080000U0000300000@0000U000030000040000R0000
30000?oooolW0000600000800000000000000P00000U000030000080001F0000;0000=D0000Q0000
jP0002h000040000`O]Ro=?k?Oc8naKl?_a6o2D0000<00000@0002P0000<00000P0002D0000<0000
100002D0000<00000@000240000800007P0001P0002PnOoo0L3ooaglool6@0006`000100002NnOoo
d?coobH0000L00000P0000H00008000000000000008U000030000080000f0000400001Olooo@o?oo
9@0000`000040000:00000`0000200009@0000`0000400009@0000`0000100008P0000`0003ooooo
9`0001P00002000000000000008000009@0000`000020000EP0002`0003N0000>P000?8000170000
10000?Kkm_`5o=3lm_^Zo73ld?`U000030000040000X000030000080000U0000300000@0000U0000
30000040000Q0000200001h0000H0000<?[oo`70ooo^nooo1T0001/0000@0000<?[ool3moolV0000
700000800006000020000000000000029@0000`000020000=P000100003VnoooL?goobD0000<0000
100002P0000<00000P0002D0000<0000100002D0000<00000@000280000<0000ooooobL0000H0000
0P00000000000002000002D0000<00000P0005H0000/0000e@0005H0003Z0000HP0000@0003?ni_m
eo]coL7kDOe0o63m9@0000`000010000:00000`0000200009@0000`0000400009@0000`000010000
8@0000P0000N00006000073iool@o?oo1_coog_noolK00004000073iool@o?oo9P0001`000020000
1P0000P00000000000000RD0000<00000P0003H0000@0000o__oog?noolU0000300000@0000X0000
30000080000U0000300000@0000U000030000040000R000030000?oooolW00006000008000000000
00000P00000U000030000080001F0000;0000=L0001n0000jP00090000040000cO]io_;kIoh1o47n
@?b`oRD0000<00000@0002P0000<00000P0002D0000<0000100002D0000<00000@00024000080000
7P0001P00000n_oo0L3ool?kool6@0006`0001000000n_ooh?koobH0000L00000P0000H000080000
00000000008U000030000080000f000040000;_koon1oooo9@0000`000040000:00000`000020000
9@0000`0000400009@0000`0000100008P0000`0003ooooo9`0001P0000200000000000000800000
9@0000`000020000EP0002`0003=0000[0000>80002i00001000093kV_n[ngcoZ_]Coa3lX?lU0000
30000040000X000030000080000U0000300000@0000U000030000040000Q0000200001h0000H0000
X?Woo`3moooFnooo2ooooa/0000@0000X?Woo`3moolV000070000080000600002000000000000002
9@0000`000020000=P000100003>nooo0oooobD0000<0000100002P0000<00000P0002D0000<0000
100002D0000<00000@000280000<0000ooooobL0000H00000P00000000000002000002D0000<0000
0P0005H0000/0000c`0009H0003R0000Z@0000@0002Mn`Wo`O_ho]3kdOh@o43o9@0000`000010000
:00000`0000200009@0000`0000400009@0000`0000100008@0000P0000N000060000:3iool1`?oo
lo_oo`I0000K000040000:3ioonPoooo9P0001`0000200001P0000P00000000000000RD0000<0000
0P0003H0000@0000jo_oog80000U0000300000@0000X000030000080000U0000300000@0000U0000
30000040000R000030000?oooolW0000600000800000000000000P00000U000030000080001F0000
;0000=D0003D0000jP000>4000040000`?^;0=_kK03Jnd<0@?b@02D0000<00000@0002P0000<0000
0P0002D0000<0000100002D0000<00000@000240000800007P0001P0000k0P000L3oooP30006@000
6`000100003b0`00L?coobH0000L00000P0000H00008000000000000008U000030000080000f0000
400003T2001`o?oo9@0000`000040000:00000`0000200009@0000`0000400009@0000`000010000
8P0000`0003ooooo9`0001P00002000000000000008000009@0000`000020000EP0002`0003K0@00
:P000>l1000g0000100005X2B_a;0W3lFP:Fo>01L?`U000030000040000X000030000080000U0000
300000@0000U000030000040000Q0000200001h0000H0000>`800070oooh0`001T0001/0000@0000
lP<00:3ooolV0000700000800006000020000000000000029@0000`000020000=P000100000i0P00
X?ooobD0000<0000100002P0000<00000P0002D0000<0000100002D0000<00000@000280000<0000
ooooobL0000H00000P00000000000002000002D0000<00000P0005H0000/0000f`400;80003_0@00
_`0000@0001J0W[oB`:PoeX2a_oP0J3o9@0000`000010000:00000`000020000B`00010000000000
1@0002D0000<00001`00P2P0000<0000100002D0000<00000000P2P0000<00000@0002D0000<0000
3@00P2P0000<00001@0000h0000D000000000100000D0000
\>"], "NumberedFigure",
 ImageSize->{464.75, 147.438},
 ImageMargins->{{7, 0}, {0, 6}},
 ImageRegion->{{0, 1}, {0, 1}}]
}, Open  ]],

Cell[CellGroupData[{

Cell["Executing Phase", "Subsection"],

Cell[TextData[{
 "The produced executable ",
 StyleBox["foomain.exe",
  FontFamily->"Courier"],
 " can be used for stand-alone execution whereas ",
 StyleBox["fooml.exe",
  FontFamily->"Courier"],
 " is used when transparent calling from Mathematica via Mathlink of the \
compiled C++ functions in fooml.exe is desired.\n"
}], "Text"],

Cell[GraphicsData["Metafile", "\<\
CF5dJ6E]HGAYHf4PEfU^I6mgLb15CDHPAVmbKF5d0@0001C40@0005P000000000000007T2001e0@00
00000000001F@@003BH00215CDH00040a1@006P000060000000000000000000080<005P2003C0000
WP0004H0000T1P0060H004M4BD<1002000<00>QRB6@0000000H000402@000o42000601/000000140
000V1Pl0603ooooo000@083hooo@nOooD@L00742000900009PH?00P0ooooo`80000>00009PH?0180
ooooo`@02P1DCU1@0`0hDGgo2P0002H63`0:05A>D5000080l0<900009PH?00P0ooooo`<0000?0000
9PH?01@0E4i@D0@0300100000@000000000500002`;@nH3h1@0000`2X@SA3PT0003j0PD000000?oo
o`0R00@0000]0@001`000?`20@00000000040000;@4100T0003j0PH020000000008R00@0000]0@80
1`000?`2000c<l`200040000;@4300D000090S<cc0840000104=00L0000K157n`Og@nH3h5@000?/2
@?l00000002@0Ol00000000BE6U]IG<PCVEg859_KF5^04@0100002d110040000;P4H00D0000:0P00
000500002@;oool21@0001@2000000@000020@406000038:4?`@n@/0001=HGAXIFeQM6USH@2P0600
=@1P05H0RP1P03H0=@1E0600100000810P070000o080003<V@8000@0000]0@D010000?010`050000
2@80c9T21`0001/4DOiA1m3i`0070000o08003<cc08000@0000]0@<010000?011@0500002@8c<l`2
1`0001/4L@:a0C00d?`40000;@4000D000090S<cc08700006`@80UH1iP0VoAD0003k0T3o00000000
T040000000004UAYKFEc84iUMb1BKfeQKP0V00@0000]0@D010000?0110040000;P4H00D0000:0P00
000500002@;oool21@0001@2ool=T`@000020@406000038:/05Po@/0001VKfm]HFU^;VEhI@100600
H02E05H0=@1P0300E@1P05D0100000810P040000;@4200D000090S<cc08700006`BAoI46D?^`0@@0
000]0@001@0000T2<c?<0PL0000K12SmP@D6o0H25@000?/2@?l00000002@0@000000000BE6U]IG<P
CVEg859_KF5^04@0100002d110040000l04500@0000^0AP01@0000X2000000D000090_ooo`850000
50;oo`fC100000810@0E0000<P[@o4022@0006I_Kfe/;VEhI@100600H02E03H0<01E0600E@040000
0P4201D0003k0T3o00000000T040000000004UAYKFEc84iUMb1BKfeQKP0V00@0000]0@D010000?01
10040000;P4H00D0000:0P00000500002@8000021@0001@2ool=T`@000020@406`00038:`?X@0Pd0
001>IFETLe^DIVm_H9AM08/0E@1E0600B`1005D0@01P0600@01E0400100000810P0500002@800002
100002d100040000;@4100@0003`0@8010000?010`0@0000n`8@00L000000;`2000000420R9CNG=d
IFd0KP@0000]0@8010000?011@0300007P0700005P@6@=co0L3JoPT0003j0PH0<0000000008R00@0
000]0@<01@0001@24?cHoPD0000C0Q3le_l40000;@4000@0000]0@4010000?010`0400009`7oo`L0
003l0P0000000P00100002d10`0500002@800002300002@3100Jol;knoh@o1[oG_`PoQ3l1@0000T2
00000P`0000T0`@0U_mNo;Go4?bFol;kT00@o0l0000V1Pl0501DCU1@100<000000000000000000T0
000V1Pl0203ooooo0@0000@0000]0@00100002d10@040000l04300<000000?oooooooooooooooooo
oooooooooooooooooooooa40000<0000200000/0000@0000NP80078100090000400007X2001b0@00
2P0001000020n?ood?Woo`T0000@0000d@h00:48000H0000300003<cc08V00007000004000060000
20000000000000029@0000`0000100009`0001P000020000000003<cc08000009@0000`000020000
:`0001P00020n?ood?Wookcmoom:o_oo5P0000`0000H0000600000`0003oool24P0000`000010000
6`0001000000000000000580001<0@000`00043oool0000000000000002@0@00o`000000019D06T0
K@1U07<0801>06D0M`0P0580K`1]0640KP00000000000000VP4000@000000000l]Soo`0000000000
0000000000000000V`4002`000000000oh;ooon2ool000000000000000000000WP4003@000000000
oh;ooon2ooooP_oooh;oo`0000000000X`4000D0000000000@000000000000000000000000000000
Y04003h000000000?`000000000000000000000000000000Y@4003h0000000001`00000000000000
0000000000000000000000000000000000000000000000000@410@410@410@009@0000`000030000
E0000900000H0000H0000<00002400000@0001XCSD3mGHe04?Wooa3lool;0000C000000000000000
0000000000000000I00004d0H@1d06P0I@1]0640M01Y06<0H@0009/0001V0000<0000600001J0000
R@000600000f0000=P0005<0001P00004P0000`0000200009`0001P0000400000000003<V@800000
9@0000`000040000:00000`000020000600000`00000c9T29@0000`000040000:`0001P000300000
d?Wood/7001:o_oo9`0001P000020000000003<cc08000009@0000`000020000:00000`000040000
600000`0000c<l`29@0000`000020000:`0001P0003Co?oo;P000:T1001[0P009P0001`000040000
1@00000000000000oooo02D0000<0000100002/0000H00009ogoonL0001?0@0000800580001<0@00
1@00043oool0000000000000002@0@0000000000019D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP00000000000000^04002`000000000oh;ooon2ool000000000000000000000^`4003P000000000
[4Nf08d_]P1/B;H0K4Rf08d_]P2/AkH0`P4003P000000000SBnf08d_]P2=;kH0SBnf08d_]P2=;kH0
b@4002h0000000000@400000000000000000000000000000c04002h0000000000000000000000000
0000000000000000d04003@000000000P0@0060300201000H0<00000000000000000000000000000
00000000000000000@410@410@410@009@0000`000050000:00000`000030000600000`0003oool2
4P0000`0000100006`000100000=ToooooooobD0000<00001@0005@0002@0000d@000581001_0@00
M@400040000J4he0oEf=@63moon`0@002`0004`0000000000000000000000000000006@0001V06l0
K`1]0640J@1^02h0I@1h06D0000l0000G`000600002F0000F@0003H0001P0000<00005@0001O0000
E0000180000<00000P0001P0000<0000<c?<0RD0000<00000@0002/0000H0000[`4004kkoon<1P00
RogoobD0000<0000100002/0000H00002@8000Oloomh1@008?gooe80001<0@000`00043oool00000
00000000002@0@0000000000019D06T0K@1U07<0801>06D0M`0P0580K`1]0640KP00000000000000
h`4000@000000000oh;oo`00000000000000000000000000i04000@000000000oh;oo`0000000000
0000000000000000iP4003`00000000000000800000000000000000000000000i`40008000000000
00000000000000000000000000000000j04000800000000000000000000000000000000000000000
j@400080000000000@00000000000000000000000000000000000000000000000000000000000000
0@410@410@410@009@0000`000030000:00000`000050000600000`0003oool24P0000`000010000
6`000100000=ToooooooobD0000<00000`0005@000240000X@400840000Q0P00Y0000040000J4he0
oEf=@402003@o?oo2@0004`000000000000000000000000000000600001V06l0K`1]06`0;P1U07P0
I@000480001P0000G`0009H0000f0000<00005@0001O0000E0000180000<00000P000580001<0@00
1@00043oool0000000000000002@0@0000000000019D06T0K@1U07<0801>06D0M`0P0580K`1]0640
KP000080000000000@00000000000000000000000000000000000080000000000000000000000000
0000000000000000n04000800000000000000000000000000000000000000000n@40008000000000
00000000000000000000000000000000nP4003@000000000@000070000000000<000000000000000
0080018000000000100000000000000000000000000000000@8001`0000000000000000000000000
00000000000000000@410@410@410@009@0000`000050000:00000`000030000600000`000000002
4P0000`0000100006`000100000=ToooooooobD0000<00001@0005@0002L0000V@4002P0001?0P00
B`000040000J4he0oEf=@1020030n_oo3@0004`0000000000000000000000000000006P0001>06D0
I@1T07<0F`0M86H0K`1_06007B1M0000RP0005@0001C0000H00004h0000l0000F@0003`0001P0000
H0000440001D0000@P000180000<00000P0002D0000<0000100002P0000<00000@0002L0000H0000
0@00004000000000000002D0000<00000@0002P0000<00000P000580001<0@000P00010000070000
00000000002l0P00000000420R9C07T0L`1d06D0K@00000000000000000000004P80008000000000
0@000000000000000000000000000000000003d0000000000@000?ooool000000000000000000000
5080048000000000SBnf00000000000000000000000000005@80048000000000SBnf000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000000
00000000000000000000000000000000000000000000000000000000000000000000000000000000
0@410@410@410@009@0000`000020000:00000`0000500009@0000`0000400009@0000`000010000
9@0000`0000200008@0000P0000N000060000=[nool1`?oog?ooo`I0000K000040000=Snool@o?oo
9P0001`0000500001P0003000000000000000RD0000<00001@0003H0000@0000e_oooa3loolU0000
300000@0000X0000300000D0000U0000300000@0000U000030000040000R000030000?oooolW0000
600000D00000000000000P00000U0000300000D0001F0000;0000?40001C00006P4006h000040000
6_o2no_n4?`Joekl8?h@o5H0000/0000;`4005<0001I0@00KP0000@0002Foekl]Ol@o9Ko`_^@013l
9@0000`000010000:00000`000050000B`000100000000001@0002D0000<00001`00P2P0000<0000
100002D0000<00000000P2P0000<00000@0002D0000<00003@00P2P0000<00000P0000h0000D0000
00000100000D0000
\>"], "Text",
 ImageSize->{345, 203.188},
 ImageMargins->{{84, 0}, {0, 0}},
 ImageRegion->{{0, 1}, {0, 1}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["MathCode Type Annotation", "Section"],

Cell["\<\
The MathCode system allows the user to associate type information with \
Mathematica variables and functions. This information is needed in order to \
generate efficient code in strongly typed languages like C++. Future versions \
of MathCode may support inference of some type information, but the current \
version requires specification of types for all variables and functions to be \
translated to C++.\
\>", "Text"],

Cell[CellGroupData[{

Cell["Basic Types", "Subsection"],

Cell["\<\
The following basic types are supported by the current version of MathCode:\
\>", "Text"],

Cell[BoxData[
 RowBox[{"Integer", ",", " ", "Real", ",", " ", "Complex"}]], "DisplayCode"],

Cell["\<\
Support for these basic types are planned, but not yet implemented:\
\>", "Text"],

Cell[BoxData[
 RowBox[{"Boolean", ",", " ", "String"}]], "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Declarations", "Subsection"],

Cell[CellGroupData[{

Cell["Global Variables", "Subsubsection"],

Cell["The type of global variables is declared as follows:", "Text"],

Cell[BoxData[
 RowBox[{"Declare", "[", "\n", "\t\t\t", 
  RowBox[{
   RowBox[{"Real", "         ", "r1"}], ",", "\n", "\t\t  ", 
   RowBox[{
    RowBox[{"Integer", "   ", "i2"}], " ", "=", " ", "3"}], ",", "\n", 
   "\t\t  ", 
   RowBox[{
    RowBox[{"Boolean", "  ", "b3"}], " ", "=", " ", "False"}], ",", 
   "\[IndentingNewLine]", "\t\t", 
   RowBox[{
    RowBox[{"Complex", " ", "c4"}], " ", "=", " ", 
    RowBox[{"3", "+", 
     RowBox[{"I", " ", "4"}]}]}]}], " ", "\n", "]"}]], "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Local Variables", "Subsubsection"],

Cell["\<\
Local variables are declared in a similar fashion, but within the body of a \
function definition:\
\>", "Text"],

Cell["\<\
...     := Module[{
\t\t\tInteger   n,
\t\t\tReal      {y,z,w},
\t\t\tReal      w2,
\t\t\tInteger   i = 1,
\t\t\tInteger   j = 0,
\t\t\tComplex   c = 1 + I 4
\t\t\t},
\t\ty = x+i+j;
\t\ty
\t\t]\
\>", "DisplayCode"],

Cell[TextData[{
 "Declared variables can be initialized by some initialization expression, \
just as in standard ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 "."
}], "Text"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Function Signatures", "Subsection"],

Cell[TextData[{
 "Function signatures of typed ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " functions are usually attached to the head of the function definitions. \
This does not change the behavior or performance of the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " function, when executed interpretively within ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", and is thus completely backwards compatible. The syntax has the following \
structure:"
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"func", "[", 
    RowBox[{
     RowBox[{"type1", " ", "x1_"}], ",", " ", "...", ",", " ", 
     RowBox[{"typen", " ", "xn_"}]}], "]"}], " ", "->", " ", "ftype"}], " ", ":=",
   " ", "..."}]], "DisplayCode"],

Cell["or for a one parameter function:", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"func", "[", 
   RowBox[{"type1", " ", "x1_"}], "]"}], "  ", ":=", "  ", 
  "..."}]], "DisplayCode"],

Cell["An example:", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"mytan", "[", 
     RowBox[{"Real", " ", "x_"}], "]"}], "->", "Real"}], " ", ":=", " ", 
   RowBox[{
    RowBox[{"Sin", "[", "x", "]"}], "/", 
    RowBox[{"Cos", "[", "x", "]"}]}]}], ";"}]], "DisplayCode"]
}, Closed]],

Cell[CellGroupData[{

Cell["Arrays", "Subsection"],

Cell[TextData[{
 "A key data structure in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " is the list structure. Nested list structures are commonly used to \
represent matrices and arrays. For example, a nested list ",
 StyleBox["{{2.1,3.1},{2.2,3.2}}",
  FontFamily->"Courier"],
 " is a two by two array of real numbers. The type of such (nested) list \
structures can be specified by array type annotations. "
}], "Text"],

Cell[TextData[{
 "It is interesting to note that ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " internally implements lists as arrays. This has the advantage of providing \
constant time indexing operations."
}], "Text"],

Cell[CellGroupData[{

Cell["Basic definition", "Subsubsection"],

Cell["Basic definitions of the variable x.", "Text"],

Cell["ArrayOf[type, {3,6,4}] x", "DisplayCode"],

Cell["or shorter", "Text"],

Cell["type[3,6,4] x", "DisplayCode"],

Cell["Example", "Example"],

Cell["A three dimensional array of real numbers:", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Real", "[", 
   RowBox[{"3", ",", "6", ",", "4"}], "]"}], " ", "x"}]], "DisplayCode"],

Cell[TextData[{
 "The sizes of array dimensions can be specified by values of integer \
variables, e.g. ",
 StyleBox["n",
  FontFamily->"Courier"],
 " and ",
 StyleBox["m",
  FontFamily->"Courier"],
 " below:"
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{"Real", "[", 
   RowBox[{"n", ",", "m"}], "]"}], " ", "x"}]], "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Unspecified Dimension Sizes", "Subsubsection"],

Cell["\<\
Typically, the sizes of arrays passed as function parameters or returned from \
a function is not known until the function is executed. Such unspecified \
array dimension sizes are indicated by the underscore (_) dimension \
placeholder. This new use of underscore is only valid within array type \
descriptors, as shown below:\
\>", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   StyleBox["type",
    FontSlant->"Italic"], "[", 
   RowBox[{"_", " ", ",", "_"}], "]"}], " ", "x_"}]], "DisplayCode"],

Cell["\<\
Named dimension placeholders are local to the scope of the type.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Array sizes in function signatures", "Subsubsection"],

Cell[TextData[{
 "Named dimension placeholders makes it possible to document array dimension \
size constraints in function signatures. For example, the following function \
signature is used for a matrix multiplication function, which multiplies two \
matrix parameters ",
 StyleBox["amat",
  FontFamily->"Courier"],
 " and ",
 StyleBox["bmat",
  FontFamily->"Courier"],
 ", with ",
 StyleBox["type",
  FontSlant->"Italic"],
 " denoting the element type."
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"MatrixMultiplication", "[", " ", 
    RowBox[{
     RowBox[{
      RowBox[{
       StyleBox["type",
        FontSlant->"Italic"], "[", 
       RowBox[{"n_", ",", "k_"}], "]"}], " ", "amat_"}], ",", 
     RowBox[{
      RowBox[{
       StyleBox["type",
        FontSlant->"Italic"], "[", 
       RowBox[{"k_", ",", "m_"}], "]"}], " ", "bmat_"}]}], "]"}], "->", 
   RowBox[{
    StyleBox["type",
     FontSlant->"Italic"], "[", 
    RowBox[{"n", ",", "m"}], "]"}]}], ":=", " ", "..."}]], "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Dimension sizes of array arguments", "Subsubsection"],

Cell[TextData[{
 "Finding dimensions size at run time, e.g. for a function array parameter \
mat, can be done by calling the standard ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " function ",
 StyleBox["Dimensions",
  FontFamily->"Courier"],
 " indexed by 1 for the size first dimension, 2 for the second dimension, \
etc. If such dimension sizes are needed for the declaration of local arrays, \
e.g. ",
 StyleBox["localmat",
  FontFamily->"Courier"],
 " below, those local arrays must be declared in an inner ",
 StyleBox["Module",
  FontFamily->"Courier"],
 " or ",
 StyleBox["Block",
  FontFamily->"Courier"],
 ", as below:"
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"(", " ", 
    RowBox[{
     RowBox[{"Real", "[", 
      RowBox[{"_", ",", "_"}], "]"}], "->", "Real"}], "\n", "\t", ")"}], " ", 
   RowBox[{"foo", "[", "mat_", "]"}]}], " ", ":=", " ", 
  RowBox[{"Module", "[", 
   RowBox[{
    RowBox[{"{", "\n", "\t\t\t\t", 
     RowBox[{
      RowBox[{
       RowBox[{"Integer", "   ", "n"}], " ", "=", " ", 
       RowBox[{
        RowBox[{"Dimensions", "[", "mat", "]"}], "\[LeftDoubleBracket]", "1", 
        "\[RightDoubleBracket]"}]}], ";", "\n", "\t\t\t\t", 
      RowBox[{
       RowBox[{"Integer", "  ", "m"}], " ", "=", " ", 
       RowBox[{
        RowBox[{"Dimensions", "[", "mat", "]"}], "\[LeftDoubleBracket]", "2", 
        "\[RightDoubleBracket]"}]}], ";"}], "\n", "\t\t\t", "}"}], ",", "\n", 
    "\t\t\t\t\t", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "\n", "                 ", 
       RowBox[{
        RowBox[{"Real", "[", 
         RowBox[{"n", ",", "m"}], "]"}], "   ", "localmat"}], "\n", 
       "\t\t\t\t", "}"}], ",", "\n", "\t\t\t\t", "..."}], "\n", "]"}]}], 
   "]"}]}]], "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "Initialization ",
 StyleBox["of arrays in declarations",
  FontSize->11.5]
}], "Subsubsection"],

Cell["\<\
Elementwise initialization of a matrix by a scalar constant (general \
expressions currently not allowed) can be done for example as below for \
locally or globally declared variables:\
\>", "Text"],

Cell["Real[2,3]  mat  = 5.0", "DisplayCode"],

Cell[TextData[{
 "which gives ",
 StyleBox["mat",
  FontFamily->"Courier"],
 " the following contents"
}], "Text"],

Cell[BoxData[
 RowBox[{"(", GridBox[{
    {"5.", "5.", "5."},
    {"5.", "5.", "5."}
   }], ")"}]], "NumberedEquation"],

Cell["Initialization by a constant matrix", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"Real", "[", 
    RowBox[{"3", ",", "3"}], "]"}], "  ", "mat"}], " ", "=", " ", 
  RowBox[{"(", GridBox[{
     {"1", "2", "3"},
     {"2", "3", "4"},
     {"3", "4", "5"}
    }], ")"}]}]], "DisplayCode"]
}, Open  ]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Compilation to C++ code", "Section"],

Cell[TextData[{
 "MathCode provides facilities to compile typed ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " functions and variables to C++ code. Functions always reside in some \
package. If no package has been specified by the user, the default package ",
 StyleBox["Global",
  FontFamily->"Courier"],
 " is usually used. The compiler is invoked by calling ",
 StyleBox["CompilePackage",
  FontFamily->"Courier"],
 ". There are essentially two ways to compile functions:"
}], "Text"],

Cell["normal compilation of the code as it is", "Itemize"],

Cell["compilation combined with symbolic evaluation", "Itemize"],

Cell["\<\
The second way is used primarily to handle symbolic operations, e.g. symbolic \
integration, simplification, substitution, etc., which may be present in the \
function body to be compiled. \
\>", "Text"],

Cell[TextData[{
 "The default method of compiling typed functions is by ",
 StyleBox["normal compilation",
  FontSlant->"Italic"],
 ". Compilation with symbolic evaluation is used for functions mentioned in \
the list of names to the optional parameter ",
 StyleBox["EvaluateFunctions",
  FontFamily->"Courier"],
 ", and should only be used for functions which contain symbolic operations."
}], "Text"],

Cell[CellGroupData[{

Cell["Calling the Code Generator", "Subsection"],

Cell[CellGroupData[{

Cell["CompilePackage", "Subsubsection"],

Cell[TextData[{
 "The code generation is started by the command ",
 StyleBox["CompilePackage",
  FontFamily->"Courier"],
 ", e.g.:"
}], "Text"],

Cell[BoxData[
 RowBox[{"CompilePackage", "[", "\"\<foo\>\"", "]"}]], "DisplayCode"],

Cell[TextData[{
 "which collects the variables and functions defined in the package context ",
 StyleBox["foo\[OpenCurlyQuote]",
  FontFamily->"Courier"],
 ", i.e. corresponding to the symbols returned by Names[\"foo`*\"] and for \
which there is a type definition."
}], "Text"],

Cell["\<\
By default MathCode compiles all typed functions and typed global variables \
within the package.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["SetCompilationOptions", "Subsubsection"],

Cell["\<\
Additional information needed to guide the compilation process can be \
specified using optional parameters to CompilePackage, or by inserting calls \
to SetCompilationOptions within the package to be compiled.\
\>", "Text"],

Cell["\<\
Below we briefly examine the different items to be compiled, and some of the \
available options.\
\>", "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compiling different items", "Subsubsection"],

Cell["Variable declarations", "Example"],

Cell[TextData[{
 "All typed global variables declared in a ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " to be compiled (e.g.  package ",
 StyleBox["foo",
  FontFamily->"Courier"],
 ") are translated to declarations in C++. Declarations are put into the \
header file ",
 StyleBox["foo.h",
  FontFamily->"Courier"],
 " and possible initialization code into the file ",
 StyleBox["foo.cc",
  FontFamily->"Courier"],
 "."
}], "Text"],

Cell["Functions", "Example"],

Cell[TextData[{
 "The default is to translate typed ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " functions into C++ without any symbolic evaluation. This produces target \
code similar to the original ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " code, i.e. loops in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " become loops in C++, ",
 StyleBox["if",
  FontFamily->"Courier"],
 "-statements are still if-statements, etc. The main difference form the \
original ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " is, e.g.,  that iterators like Do, and Table are converted to for-loops in \
C++."
}], "Text"],

Cell["Functions with symbolic operations", "Example"],

Cell[TextData[{
 "Functions which contain symbolic operations cannot be directly translated \
to C++. Fortunately, in many cases symbolic operations can be eliminated by \
symbolic expansion. In this way, symbolic operations such as symbolic \
integration, derivation, series expansion, etc. can be performed by ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " before the final code generation stage. The resulting expression, which is \
assumed to contain only non-symbolic operations, is then passed to the code \
generator which performs common subexpression elimination to speed up and \
reduce the code size before finally translating to C++ code. For this to work \
reliably, the body of the \[OpenCurlyDoubleQuote]symbolic function\
\[CloseCurlyDoubleQuote] should not contain side-effects such as assignments \
to global variables or input/output."
}], "Text"],

Cell["\<\
For example, the function below contains a symbolic series expansion and a \
symbolic substitution:\
\>", "Text"],

Cell[BoxData[
 RowBox[{" ", 
  RowBox[{
   RowBox[{
    RowBox[{
     RowBox[{"SymbSeriesSin", "[", 
      RowBox[{"Real", " ", "y_"}], "]"}], "->", "Real"}], " ", ":=", " ", 
    RowBox[{
     RowBox[{"Series", "[", 
      RowBox[{
       RowBox[{"Sin", "[", "x", "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"x", ",", "0", ",", "10"}], "}"}]}], "]"}], " ", "/.", " ", 
     RowBox[{"x", "->", "y"}]}]}], ";"}]}]], "DisplayCode"],

Cell[TextData[{
 "Therefore, the function should be symbolically evaluated before final code \
generation, and be specified using the option ",
 StyleBox["EvaluateFunctions",
  FontFamily->"Courier"],
 ":"
}], "Text"],

Cell[BoxData[
 RowBox[{"SetCompilationOptions", "[", 
  RowBox[{"EvaluateFunctions", "->", 
   RowBox[{"{", "\"\<SymbSeriesSin\>\"", "}"}]}], "]"}]], "DisplayCode"],

Cell["Main program function", "Example"],

Cell[TextData[{
 "In case a stand-alone executable should be created, the option ",
 StyleBox["MainFileAndFunction",
  FontFamily->"Courier"],
 " can be used to specify the C function ",
 StyleBox["main()",
  FontFamily->"Courier"],
 " needed in such an executable. The argument string specifies the text from \
which the file ",
 StyleBox["foomain.cc",
  FontFamily->"Courier"],
 " file is created. For example:"
}], "Text"],

Cell[BoxData[
 RowBox[{"SetCompilationOptions", "[", 
  RowBox[{"MainFileAndFunction", "->", "\"\<int main(){return 0;}\>\""}], 
  "]"}]], "DisplayCode"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Building",
 FontWeight->"Bold"]], "Subsection"],

Cell["\<\
The building process compiles all produced C++ files and links them into one \
(or two) executables.\
\>", "Text"],

Cell[BoxData[
 RowBox[{"MakeBinary", "[", "\"\<foo\>\"", "]"}]], "DisplayCode"],

Cell[TextData[{
 "The call ",
 StyleBox["MakeBinary[]",
  FontFamily->"Courier"],
 " builds all the files for both the stand-alone version of the application \
(e.g. ",
 StyleBox["foomain.exe",
  FontFamily->"Courier"],
 "), and for the interactively callable ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " version (e.g. ",
 StyleBox["fooml.exe",
  FontFamily->"Courier"],
 ")."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Installing", "Subsection"],

Cell[TextData[{
 "If compiled functions should be directly callable from within ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", the code must be installed."
}], "Text"],

Cell["The call:", "Text"],

Cell[BoxData[
 RowBox[{"InstallCode", "[", "\"\<foo\>\"", "]"}]], "DisplayCode"],

Cell[TextData[{
 "installs the binary ",
 StyleBox["fooml.exe",
  FontFamily->"Courier"],
 " into ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ". It first saves away the original interpreted functions and then creates \
function definition stubs of the compiled package in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ". This enables calling the compiled functions from within ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " via ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 "."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Executing", "Subsection"],

Cell[TextData[{
 "Functions in the compiled and installed package can be executed by standard \
function call just like functions in any standard ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " package. If stand-alone execution is desired, just run the created \
stand-alone executable."
}], "Text"]
}, Open  ]],

Cell[CellGroupData[{

Cell["UnInstalling ", "Subsection"],

Cell[TextData[{
 "When the compiled code should no longer be accessible and the ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " connection be close down, ",
 StyleBox["UnInstallCode",
  FontFamily->"Courier"],
 " should be called:"
}], "Text"],

Cell[BoxData[
 RowBox[{"UnInstallCode", "[", "\"\<foo\>\"", "]"}]], "DisplayCode"],

Cell[TextData[{
 "This will restore the original interpreted version of the package, (called ",
 StyleBox["foo",
  FontFamily->"Courier"],
 " in the example)."
}], "Text"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Matrix Operations", "Section"],

Cell[TextData[{
 "In many engineering applications, matrices and matrix manipulation are very \
common. The availability of an easy-to-use and short-handed notation for \
manipulating matrices is important for these application domains. Thus, we \
have extended the Part ([[  ]])-operation in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " to fulfill this objective. "
}], "Text"],

Cell[TextData[{
 "The current basic set of matrix operations consists of operations on array \
sections. The syntax is inspired by the syntax used by Matlab and Fortran 90, \
and is supported by the MathCode generator for 1- and 2-dimensional matrices, \
and within ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " for arbitrary number of dimensions."
}], "Text"],

Cell["\<\
Create a small matrix A containing indexed symbols of the form a[i,j]:\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"A", "=", 
   RowBox[{"Table", "[", 
    RowBox[{
     RowBox[{"a", "[", 
      RowBox[{"i", ",", "j"}], "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"i", ",", "4"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"j", ",", "5"}], "}"}]}], "]"}]}], ";", 
  RowBox[{"A", "//", "MatrixForm"}]}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "5"}], "]"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086611014938354`*^9}]
}, Open  ]],

Cell["\<\
Below we extract row 2 and 3, using the Matlab-style notation A[[2|3,_]]. \
\>", "Text"],

Cell[TextData[{
 "Compared to the standard Matlab syntax A(2:3,:) we have made a ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " compatible version by replacing colon as a binary range operator by \
vertical bar (|), and replacing colon as a placeholder for the whole range of \
a dimension by underscore (_). "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"A", "\[LeftDoubleBracket]", 
   RowBox[{"2", "|", "3"}], "\[RightDoubleBracket]"}], " ", "//", " ", 
  "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "5"}], "]"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086611015250845`*^9}]
}, Open  ]],

Cell["\<\
Extraxt all but the first two columns, using A[[_, 3|_]] which corresponds to \
standard Matlab syntax A(:,3:). The notation 3|_ here means the range from \
the 3:rd to the last column.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"A", "\[LeftDoubleBracket]", 
   RowBox[{"_", ",", 
    RowBox[{"3", "|", "_"}]}], "\[RightDoubleBracket]"}], "//", 
  "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "5"}], "]"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086611015407095`*^9}]
}, Open  ]],

Cell["Assign values to a submatrix of A:", "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{"A", "\[LeftDoubleBracket]", 
    RowBox[{
     RowBox[{"2", "|", "3"}], ",", 
     RowBox[{"2", "|", "3"}]}], "\[RightDoubleBracket]"}], " ", "=", " ", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"{", 
      RowBox[{"1", ",", "2"}], "}"}], ",", 
     RowBox[{"{", 
      RowBox[{"3", ",", "4"}], "}"}]}], "}"}]}], ";"}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{"A", " ", "//", " ", "MatrixForm"}]], "Input"],

Cell[BoxData[
 TagBox[
  RowBox[{"(", "\[NoBreak]", GridBox[{
     {
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"1", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "1"}], "]"}], "1", "2", 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"2", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "1"}], "]"}], "3", "4", 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"3", ",", "5"}], "]"}]},
     {
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "1"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "2"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "3"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "4"}], "]"}], 
      RowBox[{"a", "[", 
       RowBox[{"4", ",", "5"}], "]"}]}
    },
    GridBoxAlignment->{
     "Columns" -> {{Center}}, "ColumnsIndexed" -> {}, "Rows" -> {{Baseline}}, 
      "RowsIndexed" -> {}},
    GridBoxSpacings->{"Columns" -> {
        Offset[0.27999999999999997`], {
         Offset[0.7]}, 
        Offset[0.27999999999999997`]}, "ColumnsIndexed" -> {}, "Rows" -> {
        Offset[0.2], {
         Offset[0.4]}, 
        Offset[0.2]}, "RowsIndexed" -> {}}], "\[NoBreak]", ")"}],
  Function[BoxForm`e$, 
   MatrixForm[BoxForm`e$]]]], "Output",
 CellChangeTimes->{3.5086611015563345`*^9}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Implementing missing Mathematica functions", "Section"],

Cell[TextData[{
 "The MathCode translator directly supports a set of basic ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " functions and operations. There are still quite a number of standard ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " functions not yet included in this set. There are basically two ways to \
solve this problem: "
}], "Text"],

Cell[TextData[{
 StyleBox["Callback",
  FontSlant->"Italic"],
 ". Standard ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " functions can be made callable from external code, by providing callback \
declarations. This is easy, but often gives slow execution due to ",
 StyleBox["MathLink",
  FontSlant->"Italic"],
 " overhead and interpreted evaluation."
}], "Itemize"],

Cell[TextData[{
 StyleBox["Re-implementation",
  FontSlant->"Italic"],
 ". Standard functions can be re-implemented by hand, or by using available \
external implementations, e.g., from some library. This process is simplified \
by the availability of the system package described below."
}], "Itemize"],

Cell[CellGroupData[{

Cell["Callbacks to Mathematica", "Subsection"],

Cell[TextData[{
 "Mathematica functions can be called from outside Mathematica if they are \
made back-callable via a CallBack declaration. For example, to be able call ",
 StyleBox["RotateLeft",
  FontFamily->"Courier"],
 " in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " from generated code:"
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"RotateLeft", "[", 
     RowBox[{
      RowBox[{"Integer", "[", "_", "]"}], " ", "x_"}], "]"}], "->", 
    RowBox[{"Integer", "[", "_", "]"}]}], " ", ":=", " ", 
   RowBox[{"CallBack", "[", "]"}]}], ";", " ", 
  RowBox[{"(", 
   RowBox[{"??", "update)"}]}]}]], "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell[TextData[{
 "An example ",
 StyleBox["system.nb",
  FontSlant->"Italic"],
 " notebook"
}], "Subsection"],

Cell[TextData[{
 "This particular notebook ",
 StyleBox["system.nb",
  FontFamily->"Courier"],
 " contains an example system package (note lower-case!) with an alternative \
implementation of the standard function ",
 StyleBox["RotateLeft",
  FontFamily->"Courier"],
 ", which earlier was not in the standard subset supported by the MathCode \
translator."
}], "Text"],

Cell[CellGroupData[{

Cell["Initialization needed to use MathCode", "Subsubsection"],

Cell["Needs[\"MathCode`\"]", "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Package header", "Subsubsection"],

Cell["BeginPackage[\"system`\",{MathCodeContexts}]", "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Public Exported global symbols", "Subsubsection"],

Cell["Begin[\"system`\"]; ", "DisplayCode"],

Cell["\<\
Introduce symbols that should be exported outside the package.\
\>", "Text"],

Cell["system`RotateLeft;", "DisplayCode"],

Cell["End public section", "Text"],

Cell["End[]; ", "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Private implementation section", "Subsubsection"],

Cell["Begin[\"`Private`\"]; ", "DisplayCode"],

Cell["Define implementations of the functions and variables.", "Text"],

Cell["RotateLeft", "Example"],

Cell["Definition of RotateLeft:", "Text"],

Cell["\<\
RotateLeft[Integer[_] a_]-> Integer[_] :=  Module[{
   Integer m=Dimensions[a][[1]]
  },Module[{
   Integer[m] res
  },
  res[[1|m-1]]=a[[2|m]];
  res[[m]]=a[[1]];
  res
]]; \
\>", "DisplayCode"],

Cell["End of private section", "Example"],

Cell["End[]", "DisplayCode"],

Cell["End of package", "Example"],

Cell["EndPackage[]", "DisplayCode"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell["Compiling", "Subsection"],

Cell["Compile the system package into C++ code:", "Text"],

Cell["CompilePackage[\"system\"]", "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Building", "Subsection"],

Cell["\<\
Compile the C++ files to binaries and possibly link into an executable:\
\>", "Text"],

Cell["MakeBinary[]", "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Install and test", "Subsection"],

Cell["InstallCode[\"system\"]", "DisplayCode"],

Cell["RotateLeft test example.", "Text"],

Cell["RotateLeft[{1,2,3,4]=={2,3,4,1}", "DisplayCode"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Interfacing with external libraries", "Section"],

Cell["\<\
MathCode provides a mechanism to interface and call functions in external \
libraries and object modules which have been implemented in languages like \
Fortran, C, or C++.\
\>", "Text"],

Cell[TextData[{
 "Such functions need to be declared ",
 StyleBox["External",
  FontFamily->"Courier"],
 ", as for the Fortran subroutine ",
 StyleBox["fooext",
  FontFamily->"Courier"],
 " below, which has two input parameters and two output parameters:"
}], "Text"],

Cell[BoxData[
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"fooext", "[", 
     RowBox[{
      RowBox[{"Real", " ", "x_"}], ",", 
      RowBox[{"Integer", " ", "y_"}]}], "]"}], "->", " ", 
    RowBox[{"{", 
     RowBox[{"Real", ",", " ", "Real"}], "}"}]}], " ", ":=", " ", "\n", 
   "            ", 
   RowBox[{"External", "[", 
    RowBox[{
     RowBox[{"ParameterList", "->", 
      RowBox[{"{", 
       RowBox[{"x", ",", " ", "y", ",", " ", 
        RowBox[{"Output", " ", "u1"}], ",", " ", 
        RowBox[{"Output", " ", "u2"}]}], "}"}]}], ",", "\n", 
     "                     ", 
     RowBox[{"ExternalLanguage", "->", "\"\<Fortran\>\""}]}], "]"}]}], 
  ";"}]], "DisplayCode"],

Cell[CellGroupData[{

Cell["Linking with external libraries", "Subsection"],

Cell[TextData[{
 "The object code of external libraries need to be linked with the generated \
code to make external functions callable. Additional parameters can be \
supplied to ",
 StyleBox["MakeBinary",
  FontFamily->"Courier"],
 " for this purpose:"
}], "Text"],

Cell[BoxData[
 RowBox[{"MakeBinary", "[", 
  RowBox[{
   RowBox[{"NeedsExternalLibrary", "->", 
    RowBox[{"{", 
     RowBox[{"\"\<extlib1\>\"", ",", " ", "\"\<extlib2\>\""}], "}"}]}], ",", 
   " ", "\n", "           ", 
   RowBox[{"NeedsExternalObjectModule", "->", 
    RowBox[{"{", "\"\<file3\>\"", "}"}]}]}], " ", "]"}]], "DisplayCode"],

Cell[TextData[{
 "Note that the object module name ",
 StyleBox["file3",
  FontFamily->"Courier"],
 " in the above example would correspond e.g. to the object module named ",
 StyleBox["foo3.obj",
  FontFamily->"Courier"],
 " under Windows95, or ",
 StyleBox["foo3.o",
  FontFamily->"Courier"],
 " under Unix."
}], "Text"],

Cell[TextData[{
 "Alternatively, and usually more conveniently, options like ",
 StyleBox["NeedsExternalLibrary",
  FontFamily->"Courier"],
 " or ",
 StyleBox["NeedsExernalModule",
  FontFamily->"Courier"],
 " can be set by inserting calls to ",
 StyleBox["SetCompilationOptions",
  FontFamily->"Courier"],
 " into the package which needs to call the exernal functions, like in the \
package ",
 StyleBox["foo",
  FontFamily->"Courier"],
 " below:"
}], "Text"],

Cell[BoxData[{
 RowBox[{
  RowBox[{
   RowBox[{
    RowBox[{"BeginPackage", "[", "\"\<foo`\>\"", "]"}], "\n", "..."}], ".", 
   "\n", 
   RowBox[{"SetCompilationOptions", "[", 
    RowBox[{
     RowBox[{"NeedsExternalLibrary", "->", 
      RowBox[{"{", 
       RowBox[{"\"\<extlib1\>\"", ",", " ", "\"\<extlib2\>\""}], "}"}]}], ",",
      "\n", "                      ", 
     RowBox[{"NeedsExternalObjectModule", "->", 
      RowBox[{"{", "\"\<file3\>\"", "}"}]}]}], " ", "]"}]}], "\n", 
  "..."}], "\n", 
 RowBox[{
  RowBox[{
  "Begin", "[", "\"\<\[OpenCurlyQuote]Private\[OpenCurlyQuote]\>\"", "]"}], 
  "\n", "..."}], "\n", 
 RowBox[{"End", "[", "]"}], "\n", 
 RowBox[{"EndPackage", "[", "]"}]}], "DisplayCode"]
}, Open  ]]
}, Closed]],

Cell[CellGroupData[{

Cell["Getting Started by Examples", "Section"],

Cell[TextData[{
 "The purpose of this section is to walk through some aspects of the MathCode \
system by showing both small examples and more complete application examples, \
to help the user become acquainted with some of the type declaration and code \
generation facilities. \nSee ",
 StyleBox["readme.txt",
  FontWeight->"Bold"],
 " for current list of examples."
}], "Text"],

Cell[TextData[{
 "The performance of the generated code is also measured. The performance \
figures shown have been obtained for ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " 3.0 under WindowsNT on a Pentium 266 Mhz notebook computer. You should \
re-run these examples to obtain the correct performance figures for your \
platform."
}], "Text"],

Cell["\<\
This description is valid for execution under Sun Solaris on Sun Sparc \
workstations, Linux, and Windows95/98/NT, which are the currently supported \
platforms for code generation. \
\>", "Text"],

Cell["\<\
Other facilities, such as type declarations and Matlab-like operations, works \
on all Mathematica supported platforms.\
\>", "Text"],

Cell["\<\
To be able to run the system, there must be a valid Mathematica license on \
your computer. Also, you must have installed the MathCode application \
package. See the manual for installation description in Chapter 9. \
\>", "Text"],

Cell[CellGroupData[{

Cell["Compilation and code generation", "Subsection"],

Cell["\<\
As briefly mentioned in the introductory overview chapter, there are several \
options concerning the compilation and code generation facilities provided by \
MathCode for translation of typed Mathematica code:\
\>", "Text"],

Cell[TextData[{
 StyleBox["Target code",
  FontSlant->"Italic"],
 ". Which type of code should be produced? Currently, only C++ code \
generation is supported, but additional target languages will become \
available in the future."
}], "Itemize"],

Cell[TextData[{
 StyleBox["Compilation scope.",
  FontSlant->"Italic"],
 " Compilation of a single function, a whole package, or set of packages?"
}], "Itemize"],

Cell[TextData[{
 StyleBox["Execution integration.",
  FontSlant->"Italic"],
 " Should the compiled code be directly callable from within ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 ", or should it just be placed on some external file?"
}], "Itemize"],

Cell[TextData[{
 StyleBox["Symbolic expansion.",
  FontSlant->"Italic"],
 " Does the ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " code contain symbolic operations which should be evaluated and expanded in \
conjunction with code generation."
}], "Itemize"],

Cell["\<\
These options are explained in more detail in the manual in Chapter 7 which \
covers code generation.\
\>", "Text"],

Cell[TextData[{
 "To use typed declarations and code generation facilities for functions and \
data structures in your own package, you always need to refer to the MathCode \
application by evaluating a ",
 StyleBox["Needs",
  FontFamily->"Courier"],
 " statement:"
}], "Text"],

Cell[BoxData[
 RowBox[{"Needs", "[", "\"\<MathCode`\>\"", "]"}]], "DisplayCode"],

Cell["\<\
In order to invoke code generation functions from within your package, you \
also need to include the MathCodeLanguage context in the search path of your \
package, as below:\
\>", "Text"],

Cell[BoxData[
 RowBox[{"BeginPackage", "[", 
  RowBox[{"\"\<myPackage`\>\"", ",", 
   RowBox[{"{", "\"\<MathCodeLanguage`\>\"", "}"}]}], "]"}]], "DisplayCode"]
}, Open  ]],

Cell[CellGroupData[{

Cell["Two modes of code generation", "Subsection"],

Cell[TextData[{
 "The first example application is a rather contrived small Mathematica \
program called ",
 ButtonBox["SinSurface",
  BaseStyle->"Hyperlink",
  Active->True,
  ButtonData:>{"..\\SinSurface\\SinSurface.nb", None}],
 ", which has been designed to illustrate the two basic modes of the code \
generator: compiling without symbolic evaluation, which is performed by the \
code generation routine ",
 StyleBox["CompileFunction",
  FontFamily->"Courier"],
 " and compilation preceded by symbolic expansion, performed by ",
 StyleBox["CompileEvaluateFunction",
  FontFamily->"Courier"],
 "."
}], "Text"],

Cell[TextData[{
 StyleBox["CompileFunction",
  FontFamily->"Courier"],
 " is used used by default for generating procedural code from a typed \
Mathematica function. The function body is translated, e.g. to C++, as it is, \
without applying any symbolic transformations. Such a function may only \
contain non-symbolic operations, typically numeric computations over arrays \
and scalars. When translating to external code, e.g. in C++, emitted code \
will be rather close to the original ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " code in structure."
}], "Itemize"],

Cell[TextData[{
 StyleBox["CompileEvaluateFunction",
  FontFamily->"Courier"],
 " is used to generate code from a function that may contain symbolic \
operations, e.g. series expansion, symbolic integration, symbolic derivation, \
etc. Such symbolic operations are not meaningful to perform in languages like \
C++ or Fortran90, and are therefore performed in ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " before the final translation."
}], "Itemize"],

Cell[TextData[{
 "The result of the symbolic operations are expressions that should contain \
only non-symbolic, typically numeric, operations. This is typically the case \
since ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " always evaluates as far as possible. Thus, the function body is expanded \
(and simplified) into a usually huge symbolic expression before being \
transformed into C++ (or Fortran90) code. The result is rather unrecognizable \
compared to the original ",
 StyleBox["Mathematica",
  FontSlant->"Italic"],
 " function since both symbolic expansion and optimizations such as common \
subexpression elimination have been performed."
}], "Text"]
}, Open  ]]
}, Closed]]
},
AutoGeneratedPackage->Automatic,
ScreenStyleEnvironment->"Working",
WindowToolbars->{},
WindowSize->{934, 869},
WindowMargins->{{Automatic, 203}, {Automatic, -21}},
FrontEndVersion->"6.0 for Microsoft Windows (32-bit) (April 20, 2007)",
StyleDefinitions->Notebook[{
   Cell[
    CellGroupData[{
      Cell["Style Definitions", "Subtitle"], 
      Cell[
      "Modify the definitions below to change the default appearance of all \
cells in a given style.  Make modifications to any definition using commands \
in the Format menu.", "Text"], 
      Cell[
       CellGroupData[{
         Cell["Style Environment Names", "Section"], 
         Cell[
          StyleData[All, "Working"], PageWidth -> WindowWidth, ScriptMinSize -> 
          9], 
         Cell[
          StyleData[All, "Presentation"], PageWidth -> WindowWidth, 
          ScriptMinSize -> 12, FontSize -> 16], 
         Cell[
          StyleData[All, "Condensed"], PageWidth -> WindowWidth, 
          CellBracketOptions -> {"Margins" -> {1, 1}, "Widths" -> {0, 5}}, 
          ScriptMinSize -> 8, FontSize -> 11], 
         Cell[
          StyleData[All, "Printout"], PageWidth -> PaperWidth, ScriptMinSize -> 
          7, FontSize -> 10, 
          PrivateFontOptions -> {"FontType" -> "Outline"}]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Notebook Options", "Section"], 
         Cell[
         "The options defined for the style below will be used at the \
Notebook level.", "Text"], 
         Cell[
          StyleData["Notebook"], PageHeaders -> {{
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"], None, 
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"]}, {
             Cell[
              TextData[{
                ValueBox["FileName"]}], "Header"], None, 
             Cell[
              TextData[{
                CounterBox["Page"]}], "PageNumber"]}}, 
          PageHeaderLines -> {True, True}, 
          PrintingOptions -> {
           "FacingPages" -> True, "FirstPageHeader" -> False}, 
          CellFrameLabelMargins -> 6, StyleMenuListing -> None]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headings", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Title"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 0}, CounterIncrements -> "Title", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subtitle", 0}, {"Subsubtitle", 0}}, FontSize -> 
             34, FontColor -> GrayLevel[1], Background -> RGBColor[0, 0, 1]], 
            
            Cell[
             StyleData["Title", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 0}, FontSize -> 
             44], 
            Cell[
             StyleData["Title", "Condensed"], CellMargins -> {{0, 0}, {0, 0}},
              FontSize -> 20], 
            Cell[
             StyleData["Title", "Printout"], CellMargins -> {{0, 0}, {0, 0}}, 
             FontSize -> 24, FontTracking -> "Plain", Background -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subtitle"], ShowCellBracket -> False, 
             CellMargins -> {{0, 0}, {0, 0}}, PageBreakBelow -> False, 
             LineSpacing -> {1, 3}, ParagraphIndent -> -96, CounterIncrements -> 
             "Subtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}, {"Subsubtitle", 0}}, FontFamily -> "Helvetica", 
             FontSize -> 18, FontColor -> GrayLevel[1], Background -> 
             RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Subtitle", "Presentation"], 
             CellMargins -> {{0, 0}, {0, 0}}, LineSpacing -> {1, 6}, 
             ParagraphIndent -> -157, FontSize -> 30], 
            Cell[
             StyleData["Subtitle", "Condensed"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -78, 
             FontSize -> 14], 
            Cell[
             StyleData["Subtitle", "Printout"], 
             CellMargins -> {{0, 0}, {0, 0}}, ParagraphIndent -> -85, 
             FontSize -> 16, Background -> GrayLevel[0.6]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubtitle"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {30, 10}}, PageBreakBelow -> False, 
             CounterIncrements -> "Subsubtitle", 
             CounterAssignments -> {{"Section", 0}, {"Equation", 0}, {
               "Figure", 0}}, FontFamily -> "Helvetica", FontSize -> 14, 
             FontSlant -> "Italic"], 
            Cell[
             StyleData["Subsubtitle", "Presentation"], 
             CellMargins -> {{8, 10}, {40, 20}}, LineSpacing -> {1, 0}, 
             FontSize -> 24], 
            Cell[
             StyleData["Subsubtitle", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 8}}, FontSize -> 12], 
            Cell[
             StyleData["Subsubtitle", "Printout"], 
             CellMargins -> {{9, 10}, {50, 10}}, FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Section"], CellFrame -> {{6, 0}, {0, 1}}, CellDingbat -> 
             None, CellMargins -> {{12, Inherited}, {4, 24}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameMargins -> 6, CounterIncrements -> "Section", 
             CounterAssignments -> {{"Subsection", 0}, {"Subsubsection", 0}}, 
             FontFamily -> "Helvetica", FontSize -> 16, FontWeight -> "Bold", 
             FontColor -> RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Section", "Presentation"], 
             CellMargins -> {{10, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["Section", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Section", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, FontSize -> 14, FontTracking -> 
             "Plain", FontColor -> GrayLevel[0]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsection"], 
             CellMargins -> {{12, Inherited}, {8, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 40}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsection", "Presentation"], 
             CellMargins -> {{11, 10}, {8, 32}}, LineSpacing -> {1, 0}, 
             FontSize -> 22], 
            Cell[
             StyleData["Subsection", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["Subsection", "Printout"], 
             CellMargins -> {{9, 0}, {4, 40}}, FontSize -> 12]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Subsubsection"], CellDingbat -> "\[FilledSquare]", 
             CellMargins -> {{25, Inherited}, {8, 12}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold"], 
            Cell[
             StyleData["Subsubsection", "Presentation"], 
             CellMargins -> {{29, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Subsubsection", "Condensed"], 
             CellMargins -> {{22, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Subsubsection", "Printout"], 
             CellMargins -> {{21, 0}, {4, 20}}, FontSize -> 11]}, Open]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Body Text", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Text"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Text", FontFamily -> "Times"], 
            Cell[
             StyleData["Text", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Text", "Condensed"], CellMargins -> {{8, 10}, {4, 4}},
              LineSpacing -> {1, 1}, ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Text", "Printout"], CellMargins -> {{9, 0}, {4, 4}}, 
             ParagraphSpacing -> {0, 6}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["SmallText"], CellMargins -> {{12, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 6}, 
             CounterIncrements -> "SmallText", FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Presentation"], 
             CellMargins -> {{13, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["SmallText", "Condensed"], 
             CellMargins -> {{8, 10}, {2, 2}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["SmallText", "Printout"], 
             CellMargins -> {{9, 0}, {4, 4}}, FontSize -> 7]}, Closed]]}, 
        Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Input/Output", "Section"], 
         Cell[
         "The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Input"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, Evaluatable -> True, 
             CellGroupingRules -> "InputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             CounterIncrements -> "Input", FontWeight -> "Bold", Background -> 
             RGBColor[1, 1, 0]], 
            Cell[
             StyleData["Input", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Input", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["Input", "Printout"], CellMargins -> {{44, 0}, {6, 6}},
              Background -> GrayLevel[0.8]]}, Closed]], 
         Cell[
          StyleData["InlineInput"], Evaluatable -> True, CellGroupingRules -> 
          "InputGrouping", CellHorizontalScrolling -> True, PageBreakWithin -> 
          False, GroupPageBreakWithin -> False, DefaultFormatType -> 
          DefaultInputFormatType, AutoItalicWords -> {}, FormatType -> 
          InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
          CounterIncrements -> "Input", FontWeight -> "Bold"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Output"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, CellEditDuplicate -> True, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, 
             CellLabelMargins -> {{3, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Output", Background -> 
             RGBColor[1, 0.900008, 0.900008]], 
            Cell[
             StyleData["Output", "Presentation"], 
             CellMargins -> {{62, Inherited}, {12, 5}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Output", "Condensed"], 
             CellMargins -> {{40, Inherited}, {4, 1}}], 
            Cell[
             StyleData["Output", "Printout"], 
             CellMargins -> {{44, 0}, {6, 2}}, Background -> 
             GrayLevel[0.900008]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Message"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Message", StyleMenuListing -> 
             None, FontColor -> RGBColor[1, 0, 0]], 
            Cell[
             StyleData["Message", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Message", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Message", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Print"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "OutputGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GroupPageBreakWithin -> False, 
             GeneratedCell -> True, CellAutoOverwrite -> True, ShowCellLabel -> 
             False, DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Print", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Print", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Print", "Condensed"], 
             CellMargins -> {{50, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Print", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Graphics"], 
             CellMargins -> {{62, Inherited}, {Inherited, Inherited}}, 
             CellGroupingRules -> "GraphicsGrouping", CellHorizontalScrolling -> 
             True, PageBreakWithin -> False, GeneratedCell -> True, 
             CellAutoOverwrite -> True, ShowCellLabel -> False, 
             DefaultFormatType -> DefaultOutputFormatType, FormatType -> 
             InputForm, CounterIncrements -> "Graphics", StyleMenuListing -> 
             None], 
            Cell[
             StyleData["Graphics", "Presentation"], 
             CellMargins -> {{74, Inherited}, {Inherited, Inherited}}], 
            Cell[
             StyleData["Graphics", "Condensed"], 
             CellMargins -> {{52, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {175, 175}], 
            Cell[
             StyleData["Graphics", "Printout"], 
             CellMargins -> {{54, Inherited}, {Inherited, Inherited}}, 
             ImageSize -> {250, 250}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["CellLabel"], StyleMenuListing -> None, FontFamily -> 
             "Helvetica", FontSize -> 11, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["CellLabel", "Presentation"], FontSize -> 12], 
            Cell[
             StyleData["CellLabel", "Condensed"], FontSize -> 8], 
            Cell[
             StyleData["CellLabel", "Printout"], FontSize -> 8, FontColor -> 
             GrayLevel[0]]}, Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Unique Styles", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Author"], ShowCellBracket -> False, 
             CellMargins -> {{10, 4}, {2, 10}}, FontSize -> 16, FontSlant -> 
             "Italic"], 
            Cell[
             StyleData["Author", "Presentation"], 
             CellMargins -> {{12, 10}, {2, 12}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}, FontSize -> 20], 
            Cell[
             StyleData["Author", "Condensed"], 
             CellMargins -> {{8, 10}, {1, 4}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}, FontSize -> 12], 
            Cell[
             StyleData["Author", "Printout"], 
             CellMargins -> {{9, 0}, {4, 12}}, ParagraphSpacing -> {0, 6}, 
             FontSize -> 14]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Copyright"], ShowCellBracket -> False, 
             CellMargins -> {{10, 10}, {40, 2}}, FontFamily -> "Helvetica", 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Presentation"], 
             CellMargins -> {{12, 10}, {50, 2}}, LineSpacing -> {1, 5}, 
             FontSize -> 12], 
            Cell[
             StyleData["Copyright", "Condensed"], 
             CellMargins -> {{8, 10}, {12, 1}}, LineSpacing -> {1, 2}, 
             FontSize -> 9], 
            Cell[
             StyleData["Copyright", "Printout"], 
             CellMargins -> {{9, 0}, {72, 4}}, FontSize -> 7]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Example"], CellMargins -> {{12, 10}, {5, 12}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 12}, 
             CounterIncrements -> "Example", FontFamily -> "Times", 
             FontWeight -> "Bold"], 
            Cell[
             StyleData["Example", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 20}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["Example", "Condensed"], 
             CellMargins -> {{8, 10}, {4, 8}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["Example", "Printout"], 
             CellMargins -> {{9, 0}, {4, 10}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Definition"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Definition", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> RGBColor[0.2, 0.700008, 0.700008]], 
            Cell[
             StyleData["Definition", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Definition", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Definition", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.6]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Theorem"], CellFrame -> {{3, 0}, {0, 0}}, 
             CellMargins -> {{52, 10}, {8, 8}}, PageBreakWithin -> False, 
             GroupPageBreakWithin -> False, 
             CellLabelMargins -> {{23, Inherited}, {Inherited, Inherited}}, 
             ShowStringCharacters -> True, CounterIncrements -> "Theorem", 
             FontFamily -> "Helvetica", FontWeight -> "Bold", FontColor -> 
             GrayLevel[1], Background -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Theorem", "Presentation"], 
             CellMargins -> {{62, Inherited}, {5, 12}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["Theorem", "Condensed"], 
             CellMargins -> {{40, 10}, {1, 4}}], 
            Cell[
             StyleData["Theorem", "Printout"], 
             CellMargins -> {{44, 0}, {2, 6}}, Background -> GrayLevel[0.4]]},
            Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseMain"], CellFrame -> {{6, 0}, {0, 1}}, 
             CellMargins -> {{12, Inherited}, {4, 20}}, 
             CellGroupingRules -> {"SectionGrouping", 30}, PageBreakBelow -> 
             False, CellFrameColor -> RGBColor[0.571389, 0.19675, 0.570504], 
             CounterIncrements -> "Subsection", 
             CounterAssignments -> {{"Subsubsection", 0}}, FontFamily -> 
             "Helvetica", FontSize -> 16, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["ExerciseMain", "Presentation"], 
             CellMargins -> {{18, 10}, {8, 32}}, LineSpacing -> {1, 2}, 
             FontSize -> 24, FontTracking -> "Condensed"], 
            Cell[
             StyleData["ExerciseMain", "Condensed"], 
             CellMargins -> {{8, Inherited}, {2, 12}}, FontSize -> 12], 
            Cell[
             StyleData["ExerciseMain", "Printout"], 
             CellMargins -> {{9, 0}, {2, 50}}, CellFrameColor -> 
             GrayLevel[0.500008], FontSize -> 14, FontTracking -> "Plain", 
             FontColor -> GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Exercise"], CellDingbat -> "\[FilledDownTriangle]", 
             CellMargins -> {{23, Inherited}, {4, 18}}, 
             CellGroupingRules -> {"SectionGrouping", 50}, PageBreakBelow -> 
             False, CounterIncrements -> "Subsubsection", FontFamily -> 
             "Times", FontSize -> 13, FontWeight -> "Bold", FontColor -> 
             RGBColor[0.571389, 0.19675, 0.570504]], 
            Cell[
             StyleData["Exercise", "Presentation"], 
             CellMargins -> {{33, 10}, {8, 26}}, LineSpacing -> {1, 0}, 
             FontSize -> 18], 
            Cell[
             StyleData["Exercise", "Condensed"], 
             CellMargins -> {{17, Inherited}, {2, 12}}, FontSize -> 10], 
            Cell[
             StyleData["Exercise", "Printout"], 
             CellFrame -> {{0, 0}, {0.5, 0}}, CellDingbat -> None, 
             CellMargins -> {{9, 0}, {6, 20}}, FontSize -> 11, FontColor -> 
             GrayLevel[0]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ExerciseText"], CellMargins -> {{62, 10}, {5, 5}}, 
             LineSpacing -> {1, 3}, ParagraphSpacing -> {0, 8}, 
             CounterIncrements -> "ExcersiceText", FontFamily -> "Times"], 
            Cell[
             StyleData["ExerciseText", "Presentation"], 
             CellMargins -> {{74, 10}, {8, 8}}, LineSpacing -> {1, 5}, 
             ParagraphSpacing -> {0, 12}], 
            Cell[
             StyleData["ExerciseText", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 2}}, LineSpacing -> {1, 1}, 
             ParagraphSpacing -> {0, 4}], 
            Cell[
             StyleData["ExerciseText", "Printout"], 
             CellMargins -> {{54, 0}, {4, 4}}, ParagraphSpacing -> {0, 6}]}, 
           Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Itemize"], CellDingbat -> "\[FilledSmallCircle]", 
             CellMargins -> {{65, Inherited}, {8, 8}}, FontFamily -> "Times", 
             FontSize -> 13, FontWeight -> "Plain", FontSlant -> "Plain", 
             FontTracking -> "Plain", 
             FontVariations -> {
              "Outline" -> False, "Shadow" -> False, "Underline" -> False}], 
            Cell[
             StyleData["Itemize", "Presentation"], 
             CellMargins -> {{Inherited, 10}, {8, 12}}, LineSpacing -> {1, 0},
              FontSize -> 18], 
            Cell[
             StyleData["Itemize", "Condensed"], 
             CellMargins -> {{Inherited, Inherited}, {2, 12}}, FontSize -> 
             10], 
            Cell[
             StyleData["Itemize", "Printout"], 
             CellMargins -> {{Inherited, 0}, {4, 20}}, FontSize -> 11]}, 
           Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayCode"], CellFrame -> 2, 
             CellMargins -> {{52, 10}, {8, 8}}, CellHorizontalScrolling -> 
             True, PageBreakWithin -> True, 
             CellLabelMargins -> {{5, Inherited}, {Inherited, Inherited}}, 
             DefaultFormatType -> DefaultInputFormatType, FormatType -> 
             InputForm, ShowStringCharacters -> True, NumberMarks -> True, 
             FontWeight -> "Bold", Background -> RGBColor[1, 1, 0]], 
            Cell[
             StyleData["DisplayCode", "Presentation"], 
             CellMargins -> {{62, Inherited}, {10, 10}}, 
             LineSpacing -> {1, 0}], 
            Cell[
             StyleData["DisplayCode", "Condensed"], 
             CellMargins -> {{40, 10}, {4, 4}}], 
            Cell[
             StyleData["DisplayCode", "Printout"], 
             CellMargins -> {{44, 0}, {6, 6}}, Background -> GrayLevel[0.8]]},
            Closed]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Formulas and Programming", "Section"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["DisplayFormula"], CellMargins -> {{62, 10}, {2, 10}}, 
             CellHorizontalScrolling -> True, ScriptLevel -> 0, 
             SingleLetterItalics -> True, 
             UnderoverscriptBoxOptions -> {LimitsPositioning -> True}], 
            Cell[
             StyleData["DisplayFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 18], 
            Cell[
             StyleData["DisplayFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["DisplayFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["ChemicalFormula"], CellMargins -> {{62, 10}, {2, 10}},
              AutoSpacing -> False, ScriptBaselineShifts -> {0.6, Automatic}, 
             ScriptLevel -> 1, SingleLetterItalics -> False, ZeroWidthTimes -> 
             True], 
            Cell[
             StyleData["ChemicalFormula", "Presentation"], 
             CellMargins -> {{74, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Condensed"], 
             CellMargins -> {{52, 10}, {2, 10}}, FontSize -> 10], 
            Cell[
             StyleData["ChemicalFormula", "Printout"], 
             CellMargins -> {{54, 10}, {2, 10}}, FontSize -> 10]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Program"], CellMargins -> {{12, 10}, {Inherited, 6}}, 
             FontFamily -> "Courier"], 
            Cell[
             StyleData["Program", "Presentation"], 
             CellMargins -> {{13, 30}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Condensed"], 
             CellMargins -> {{8, 10}, {Inherited, 4}}, FontSize -> 9.5], 
            Cell[
             StyleData["Program", "Printout"], 
             CellMargins -> {{9, 0}, {Inherited, 4}}, FontSize -> 9.5]}, 
           Closed]]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Automatic Numbering", "Section"], 
         Cell[
         "The following styles are useful for numbered equations, figures, \
etc.  They automatically give the cell a FrameLabel containing a reference to \
a particular counter, and also increment that counter.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedEquation"], 
             CellMargins -> {{62, 10}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, 
                Cell[
                 TextData[{"(", 
                   CounterBox["NumberedEquation"], ")"}]]}, {None, None}}, 
             DefaultFormatType -> DefaultInputFormatType, CounterIncrements -> 
             "NumberedEquation", FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedEquation", "Presentation"], 
             CellMargins -> {{74, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Condensed"], 
             CellMargins -> {{52, 10}, {Inherited, Inherited}}], 
            Cell[
             StyleData["NumberedEquation", "Printout"], 
             CellMargins -> {{54, 0}, {Inherited, Inherited}}]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedFigure"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Figure ", 
                   CounterBox["NumberedFigure"]}]], None}}, CounterIncrements -> 
             "NumberedFigure", 
             ImageMargins -> {{43, Inherited}, {Inherited, 0}}, 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedFigure", "Presentation"]], 
            Cell[
             StyleData["NumberedFigure", "Condensed"]], 
            Cell[
             StyleData["NumberedFigure", "Printout"]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["NumberedTable"], 
             CellMargins -> {{62, 145}, {Inherited, Inherited}}, 
             CellFrameLabels -> {{None, None}, {
                Cell[
                 TextData[{"Table ", 
                   CounterBox["NumberedTable"]}]], None}}, TextAlignment -> 
             Center, CounterIncrements -> "NumberedTable", 
             FormatTypeAutoConvert -> False], 
            Cell[
             StyleData["NumberedTable", "Presentation"]], 
            Cell[
             StyleData["NumberedTable", "Condensed"]], 
            Cell[
             StyleData["NumberedTable", "Printout"]]}, Open]]}, Open]], 
      Cell[
       CellGroupData[{
         Cell["Styles for Headers and Footers", "Section"], 
         Cell[
          StyleData["Header"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontSlant -> "Italic"], 
         Cell[
          StyleData["Footer"], CellMargins -> {{0, 0}, {0, 4}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 6],
          
         Cell[
          StyleData["PageNumber"], CellMargins -> {{0, 0}, {4, 1}}, 
          StyleMenuListing -> None, FontFamily -> "Helvetica", FontSize -> 9, 
          FontWeight -> "Bold"]}, Closed]], 
      Cell[
       CellGroupData[{
         Cell["Hyperlink Styles", "Section"], 
         Cell[
         "The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["Hyperlink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, 
             FontVariations -> {"Underline" -> True}, FontColor -> 
             RGBColor[0, 0, 1], Background -> None, 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`NotebookLocate[#2]}]& ), ButtonNote -> 
               ButtonData}], 
            Cell[
             StyleData["Hyperlink", "Presentation"]], 
            Cell[
             StyleData["Hyperlink", "Condensed"]], 
            Cell[
             StyleData["Hyperlink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Open]], 
         Cell[
         "The following styles are for linking automatically to the on-line \
help system.", "Text"], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["MainBookLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["MainBook", #]}]& )}], 
            Cell[
             StyleData["MainBookLink", "Presentation"]], 
            Cell[
             StyleData["MainBookLink", "Condensed"]], 
            Cell[
             StyleData["MainBookLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Open]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["AddOnsLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["AddOns", #]}]& )}], 
            Cell[
             StyleData["AddOnsLink", "Presentation"]], 
            Cell[
             StyleData["AddOnsLink", "Condensed"]], 
            Cell[
             StyleData["AddOnLink", "Printout"], FontColor -> GrayLevel[0], 
             Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["RefGuideLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontFamily -> "Courier", 
             FontColor -> GrayLevel[1], Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["RefGuideLink", #]}]& )}], 
            Cell[
             StyleData["RefGuideLink", "Presentation"]], 
            Cell[
             StyleData["RefGuideLink", "Condensed"]], 
            Cell[
             StyleData["RefGuideLink", "Printout"], FontColor -> GrayLevel[0],
              Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["GettingStartedLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["GettingStarted", #]}]& )}], 
            Cell[
             StyleData["GettingStartedLink", "Presentation"]], 
            Cell[
             StyleData["GettingStartedLink", "Condensed"]], 
            Cell[
             StyleData["GettingStartedLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]], 
         Cell[
          CellGroupData[{
            Cell[
             StyleData["OtherInformationLink"], StyleMenuListing -> None, 
             ButtonStyleMenuListing -> Automatic, FontColor -> GrayLevel[1], 
             Background -> RGBColor[1, 0.4, 0], 
             ButtonBoxOptions -> {
              Active -> True, Appearance -> {Automatic, None}, 
               ButtonFunction :> (FrontEndExecute[{
                  FrontEnd`HelpBrowserLookup["OtherInformation", #]}]& )}], 
            Cell[
             StyleData["OtherInformationLink", "Presentation"]], 
            Cell[
             StyleData["OtherInformationLink", "Condensed"]], 
            Cell[
             StyleData["OtherInformationLink", "Printout"], FontColor -> 
             GrayLevel[0], Background -> GrayLevel[1]]}, Closed]]}, Closed]]},
      Open]]}, Visible -> False, FrontEndVersion -> 
  "6.0 for Microsoft Windows (32-bit) (April 20, 2007)", StyleDefinitions -> 
  "Default.nb"]
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[580, 21, 38, 0, 55, "Title"],
Cell[621, 23, 60, 0, 54, "Copyright"],
Cell[CellGroupData[{
Cell[706, 27, 31, 0, 54, "Section"],
Cell[740, 29, 88, 2, 29, "Text"],
Cell[831, 33, 124, 3, 33, "Itemize"],
Cell[958, 38, 72, 0, 33, "Itemize"],
Cell[1033, 40, 133, 3, 33, "Itemize"],
Cell[1169, 45, 114, 3, 33, "Itemize"],
Cell[1286, 50, 255, 5, 48, "Text"]
}, Closed]],
Cell[CellGroupData[{
Cell[1578, 60, 38, 0, 34, "Section"],
Cell[1619, 62, 125, 3, 25, "Text"],
Cell[1747, 67, 65, 1, 48, "Input",
 InitializationCell->True],
Cell[1815, 70, 236, 6, 25, "Text"],
Cell[2054, 78, 144, 3, 52, "DisplayCode"],
Cell[CellGroupData[{
Cell[2223, 85, 43, 0, 44, "Subsection"],
Cell[2269, 87, 902, 24, 79, "Text"],
Cell[3174, 113, 12752, 160, 257, "NumberedFigure"]
}, Open  ]],
Cell[CellGroupData[{
Cell[15963, 278, 36, 0, 44, "Subsection"],
Cell[16002, 280, 336, 9, 43, "Text"],
Cell[16341, 291, 16484, 206, 185, "NumberedFigure"]
}, Open  ]],
Cell[CellGroupData[{
Cell[32862, 502, 37, 0, 44, "Subsection"],
Cell[32902, 504, 334, 9, 73, "Text"],
Cell[33239, 515, 7368, 94, 214, "Text"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[40656, 615, 43, 0, 54, "Section"],
Cell[40702, 617, 431, 7, 67, "Text"],
Cell[CellGroupData[{
Cell[41158, 628, 33, 0, 46, "Subsection"],
Cell[41194, 630, 99, 2, 29, "Text"],
Cell[41296, 634, 90, 1, 54, "DisplayCode"],
Cell[41389, 637, 91, 2, 29, "Text"],
Cell[41483, 641, 71, 1, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[41591, 647, 34, 0, 46, "Subsection"],
Cell[CellGroupData[{
Cell[41650, 651, 41, 0, 37, "Subsubsection"],
Cell[41694, 653, 68, 0, 29, "Text"],
Cell[41765, 655, 499, 13, 154, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[42301, 673, 40, 0, 37, "Subsubsection"],
Cell[42344, 675, 122, 3, 29, "Text"],
Cell[42469, 680, 224, 12, 234, "DisplayCode"],
Cell[42696, 694, 181, 6, 29, "Text"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[42926, 706, 41, 0, 30, "Subsection"],
Cell[42970, 708, 488, 13, 48, "Text"],
Cell[43461, 723, 252, 7, 54, "DisplayCode"],
Cell[43716, 732, 48, 0, 29, "Text"],
Cell[43767, 734, 134, 4, 54, "DisplayCode"],
Cell[43904, 740, 27, 0, 29, "Text"],
Cell[43934, 742, 264, 8, 54, "DisplayCode"]
}, Closed]],
Cell[CellGroupData[{
Cell[44235, 755, 28, 0, 30, "Subsection"],
Cell[44266, 757, 431, 10, 48, "Text"],
Cell[44700, 769, 228, 6, 29, "Text"],
Cell[CellGroupData[{
Cell[44953, 779, 41, 0, 37, "Subsubsection"],
Cell[44997, 781, 52, 0, 29, "Text"],
Cell[45052, 783, 47, 0, 54, "DisplayCode"],
Cell[45102, 785, 26, 0, 29, "Text"],
Cell[45131, 787, 36, 0, 54, "DisplayCode"],
Cell[45170, 789, 26, 0, 36, "Example"],
Cell[45199, 791, 58, 0, 29, "Text"],
Cell[45260, 793, 120, 3, 54, "DisplayCode"],
Cell[45383, 798, 221, 9, 29, "Text"],
Cell[45607, 809, 110, 3, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[45754, 817, 52, 0, 37, "Subsubsection"],
Cell[45809, 819, 351, 6, 48, "Text"],
Cell[46163, 827, 155, 5, 54, "DisplayCode"],
Cell[46321, 834, 88, 2, 29, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[46446, 841, 59, 0, 37, "Subsubsection"],
Cell[46508, 843, 468, 14, 48, "Text"],
Cell[46979, 859, 547, 18, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[47563, 882, 59, 0, 37, "Subsubsection"],
Cell[47625, 884, 649, 20, 67, "Text"],
Cell[48277, 906, 1116, 30, 234, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[49430, 941, 113, 4, 37, "Subsubsection"],
Cell[49546, 947, 208, 4, 29, "Text"],
Cell[49757, 953, 44, 0, 54, "DisplayCode"],
Cell[49804, 955, 114, 5, 29, "Text"],
Cell[49921, 962, 119, 4, 37, "NumberedEquation"],
Cell[50043, 968, 51, 0, 29, "Text"],
Cell[50097, 970, 249, 9, 81, "DisplayCode"]
}, Open  ]]
}, Closed]]
}, Closed]],
Cell[CellGroupData[{
Cell[50407, 986, 42, 0, 34, "Section"],
Cell[50452, 988, 495, 12, 48, "Text"],
Cell[50950, 1002, 58, 0, 33, "Itemize"],
Cell[51011, 1004, 64, 0, 33, "Itemize"],
Cell[51078, 1006, 213, 4, 48, "Text"],
Cell[51294, 1012, 402, 9, 48, "Text"],
Cell[CellGroupData[{
Cell[51721, 1025, 48, 0, 46, "Subsection"],
Cell[CellGroupData[{
Cell[51794, 1029, 39, 0, 37, "Subsubsection"],
Cell[51836, 1031, 143, 5, 29, "Text"],
Cell[51982, 1038, 83, 1, 54, "DisplayCode"],
Cell[52068, 1041, 278, 6, 48, "Text"],
Cell[52349, 1049, 121, 3, 29, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[52507, 1057, 46, 0, 37, "Subsubsection"],
Cell[52556, 1059, 234, 4, 48, "Text"],
Cell[52793, 1065, 121, 3, 29, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[52951, 1073, 50, 0, 37, "Subsubsection"],
Cell[53004, 1075, 40, 0, 36, "Example"],
Cell[53047, 1077, 440, 15, 48, "Text"],
Cell[53490, 1094, 28, 0, 36, "Example"],
Cell[53521, 1096, 640, 20, 67, "Text"],
Cell[54164, 1118, 53, 0, 36, "Example"],
Cell[54220, 1120, 874, 14, 105, "Text"],
Cell[55097, 1136, 123, 3, 29, "Text"],
Cell[55223, 1141, 435, 13, 54, "DisplayCode"],
Cell[55661, 1156, 217, 6, 29, "Text"],
Cell[55881, 1164, 164, 3, 54, "DisplayCode"],
Cell[56048, 1169, 40, 0, 36, "Example"],
Cell[56091, 1171, 425, 12, 48, "Text"],
Cell[56519, 1185, 153, 3, 54, "DisplayCode"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[56721, 1194, 71, 1, 46, "Subsection"],
Cell[56795, 1197, 124, 3, 29, "Text"],
Cell[56922, 1202, 79, 1, 54, "DisplayCode"],
Cell[57004, 1205, 395, 15, 48, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[57436, 1225, 32, 0, 46, "Subsection"],
Cell[57471, 1227, 176, 5, 29, "Text"],
Cell[57650, 1234, 25, 0, 29, "Text"],
Cell[57678, 1236, 80, 1, 54, "DisplayCode"],
Cell[57761, 1239, 517, 18, 48, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[58315, 1262, 31, 0, 46, "Subsection"],
Cell[58349, 1264, 306, 7, 48, "Text"]
}, Open  ]],
Cell[CellGroupData[{
Cell[58692, 1276, 35, 0, 46, "Subsection"],
Cell[58730, 1278, 247, 8, 29, "Text"],
Cell[58980, 1288, 82, 1, 54, "DisplayCode"],
Cell[59065, 1291, 171, 5, 29, "Text"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[59285, 1302, 36, 0, 34, "Section"],
Cell[59324, 1304, 388, 8, 43, "Text"],
Cell[59715, 1314, 369, 8, 43, "Text"],
Cell[60087, 1324, 94, 2, 25, "Text"],
Cell[CellGroupData[{
Cell[60206, 1330, 337, 11, 48, "Input"],
Cell[60546, 1343, 1923, 60, 105, "Output"]
}, Open  ]],
Cell[62484, 1406, 98, 2, 25, "Text"],
Cell[62585, 1410, 330, 7, 43, "Text"],
Cell[CellGroupData[{
Cell[62940, 1421, 163, 4, 48, "Input"],
Cell[63106, 1427, 1259, 38, 76, "Output"]
}, Open  ]],
Cell[64380, 1468, 209, 4, 25, "Text"],
Cell[CellGroupData[{
Cell[64614, 1476, 178, 5, 48, "Input"],
Cell[64795, 1483, 1403, 44, 105, "Output"]
}, Open  ]],
Cell[66213, 1530, 50, 0, 25, "Text"],
Cell[66266, 1532, 377, 12, 48, "Input"],
Cell[CellGroupData[{
Cell[66668, 1548, 69, 1, 48, "Input"],
Cell[66740, 1551, 1683, 52, 105, "Output"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[68472, 1609, 61, 0, 54, "Section"],
Cell[68536, 1611, 361, 9, 67, "Text"],
Cell[68900, 1622, 375, 11, 67, "Itemize"],
Cell[69278, 1635, 303, 6, 67, "Itemize"],
Cell[CellGroupData[{
Cell[69606, 1645, 46, 0, 46, "Subsection"],
Cell[69655, 1647, 316, 9, 48, "Text"],
Cell[69974, 1658, 334, 10, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[70345, 1673, 109, 5, 46, "Subsection"],
Cell[70457, 1680, 368, 10, 67, "Text"],
Cell[CellGroupData[{
Cell[70850, 1694, 62, 0, 37, "Subsubsection"],
Cell[70915, 1696, 43, 0, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[70995, 1701, 39, 0, 37, "Subsubsection"],
Cell[71037, 1703, 67, 0, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[71141, 1708, 55, 0, 37, "Subsubsection"],
Cell[71199, 1710, 43, 0, 54, "DisplayCode"],
Cell[71245, 1712, 86, 2, 29, "Text"],
Cell[71334, 1716, 41, 0, 54, "DisplayCode"],
Cell[71378, 1718, 34, 0, 29, "Text"],
Cell[71415, 1720, 30, 0, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[71482, 1725, 55, 0, 37, "Subsubsection"],
Cell[71540, 1727, 45, 0, 54, "DisplayCode"],
Cell[71588, 1729, 70, 0, 29, "Text"],
Cell[71661, 1731, 29, 0, 36, "Example"],
Cell[71693, 1733, 41, 0, 29, "Text"],
Cell[71737, 1735, 205, 10, 198, "DisplayCode"],
Cell[71945, 1747, 41, 0, 36, "Example"],
Cell[71989, 1749, 28, 0, 54, "DisplayCode"],
Cell[72020, 1751, 33, 0, 36, "Example"],
Cell[72056, 1753, 35, 0, 54, "DisplayCode"]
}, Open  ]]
}, Open  ]],
Cell[CellGroupData[{
Cell[72140, 1759, 31, 0, 46, "Subsection"],
Cell[72174, 1761, 57, 0, 29, "Text"],
Cell[72234, 1763, 49, 0, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[72320, 1768, 30, 0, 46, "Subsection"],
Cell[72353, 1770, 95, 2, 29, "Text"],
Cell[72451, 1774, 35, 0, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[72523, 1779, 38, 0, 46, "Subsection"],
Cell[72564, 1781, 46, 0, 54, "DisplayCode"],
Cell[72613, 1783, 40, 0, 29, "Text"],
Cell[72656, 1785, 54, 0, 54, "DisplayCode"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[72759, 1791, 54, 0, 34, "Section"],
Cell[72816, 1793, 196, 4, 48, "Text"],
Cell[73015, 1799, 267, 8, 48, "Text"],
Cell[73285, 1809, 684, 20, 94, "DisplayCode"],
Cell[CellGroupData[{
Cell[73994, 1833, 53, 0, 46, "Subsection"],
Cell[74050, 1835, 266, 7, 48, "Text"],
Cell[74319, 1844, 341, 8, 74, "DisplayCode"],
Cell[74663, 1854, 322, 11, 48, "Text"],
Cell[74988, 1867, 460, 15, 67, "Text"],
Cell[75451, 1884, 715, 20, 234, "DisplayCode"]
}, Open  ]]
}, Closed]],
Cell[CellGroupData[{
Cell[76215, 1910, 46, 0, 34, "Section"],
Cell[76264, 1912, 380, 8, 79, "Text"],
Cell[76647, 1922, 354, 8, 48, "Text"],
Cell[77004, 1932, 206, 4, 29, "Text"],
Cell[77213, 1938, 143, 3, 29, "Text"],
Cell[77359, 1943, 240, 4, 48, "Text"],
Cell[CellGroupData[{
Cell[77624, 1951, 53, 0, 46, "Subsection"],
Cell[77680, 1953, 234, 4, 48, "Text"],
Cell[77917, 1959, 246, 6, 50, "Itemize"],
Cell[78166, 1967, 161, 4, 33, "Itemize"],
Cell[78330, 1973, 259, 7, 33, "Itemize"],
Cell[78592, 1982, 266, 8, 50, "Itemize"],
Cell[78861, 1992, 125, 3, 29, "Text"],
Cell[78989, 1997, 277, 7, 48, "Text"],
Cell[79269, 2006, 80, 1, 54, "DisplayCode"],
Cell[79352, 2009, 198, 4, 29, "Text"],
Cell[79553, 2015, 159, 3, 54, "DisplayCode"]
}, Open  ]],
Cell[CellGroupData[{
Cell[79749, 2023, 50, 0, 46, "Subsection"],
Cell[79802, 2025, 613, 16, 67, "Text"],
Cell[80418, 2043, 578, 12, 68, "Itemize"],
Cell[80999, 2057, 460, 10, 68, "Itemize"],
Cell[81462, 2069, 674, 14, 86, "Text"]
}, Open  ]]
}, Closed]]
}
]
*)

(* End of internal cache information *)
